import pygame
import time
import math
import os
import random


_fonts = {}
def get_font(size):
	global _fonts
	key = 'f' + str(size)
	font = _fonts.get(key)
	if font == None:
		font = pygame.font.SysFont(pygame.font.get_default_font(), size)
		_fonts[key] = font
	return font

_text = {}
def get_text(text, size, color):
	global _text
	key = str(size) + "," + str(color) + "|" + text
	image = _text.get(key)
	if image == None:
		font = get_font(size)
		image = font.render(text, True, color)
		_text[key] = image
	return image

_image_library = {}

def get_image(path):
	global _image_library
	image = _image_library.get(path)
	if image == None:
		image = pygame.image.load(canonicalize_path('images/' + path))
		_image_library[path] = image
	return image



class MyEvent:
	def __init__(self, key, down):
		self.key = key
		self.down = down
		self.up = not down

_input_manager = None

def get_input_manager():
	global _input_manager
	if _input_manager == None:
		_input_manager = InputManager()
	return _input_manager

class InputManager:
	def __init__(self):
		self.cursor = (0, 0)
		self.mouse_pressed = False
		self.mouse_events = []
		self.joysticks = []
		self.active_joystick = -1
		self.raw_keyups = []
		self.read_config_save()
		self.active_actual_joystick = -1
		self.activate_joysticks()
		self.events = []
		self.quitAttempt = False
		self.init_default_key_config()
		self.my_pressed = {
			'start': False,
			'left': False,
			'right': False,
			'up': False,
			'down': False,
			'spray': False,
			'walkie': False
		}
		
		self.axes = [0.0, 0.0]
	
	def init_default_key_config(self):
		self._key_mapping = {
			pygame.K_RETURN: 'start',
			pygame.K_LEFT: 'left',
			pygame.K_RIGHT: 'right',
			pygame.K_UP: 'up',
			pygame.K_DOWN: 'down',
			pygame.K_SPACE: 'spray',
			pygame.K_w: 'walkie'
		}
		t = read_file('data/key_config.txt')
		
		things = 'start left right up down spray walkie'.split()
		if t != None:
			for line in t.split('\n'):
				parts = line.split(':')
				if len(parts) == 2:
					action = parts[0]
					value = parseInt(parts[1])
					
					if value > 0 and action in things:
						for k in self._key_mapping.keys():
							if self._key_mapping[k] == action:
								self._key_mapping.pop(k)
								break
						self._key_mapping[value] = action
		
	
	def save_key_config(self):
		output = []
		for key in self._key_mapping.keys():
			action = self._key_mapping[key]
			output.append(action + ':' + str(key))
		output = '\r\n'.join(output)
		write_file('data/key_config.txt', output)
	
	def set_key_config(self, action, key):
		for k in self._key_mapping.keys():
			if self._key_mapping[k] == action:
				self._key_mapping.pop(k)
				break
		self._key_mapping[key] = action
	
	def set_active_actual_joystick(self, id):
		if id == -1:
			self.active_actual_joystick = -1
			self.active_joystick = -1
		elif len(self.actual_joysticks) > id:
			self.active_actual_joystick = id
			name = trim(self.actual_joysticks[self.active_actual_joystick].get_name())
			found = False
			i = 0
			for js in self.joysticks:
				if js.get('name', '').lower() == name.lower():
					found = True
					self.active_joystick = i
				i += 1
			if not found:
				self.active_joystick = len(self.joysticks)
				self.joysticks.append({'name': name})
	
	def get_config_label_for_key_for_keyboard(self, key):
		for k in self._key_mapping.keys():
			if self._key_mapping[k] == key:
				output = pygame.key.name(k)
				return output[0].upper() + output[1:]
		return "Not configured!"
	
	def get_config_label_for_key_for_active(self, key):
		if self.active_joystick != -1:
			js = self.joysticks[self.active_joystick]
			value = js.get(key, None)
			if value == None:
				return "Not configured!"
			if value[0] == 'axis':
				output = "Axis " + str(value[1])
				if value[2].endswith('+'):
					output = "Positive " + output
				else:
					output = "Negative " + output
				
			elif value[0] == 'hat':
				output = "Hat " + str(value[1])
				if value[2].startswith('x'):
					output = "X " + output
				else:
					output = "Y " + output
				if value[2].endswith('-'):
					output = "Negative " + output
				else:
					output = "Positive " + output
			else:
				output = "Button " + str(value[1])
				
			return output
		return "---"
	
	def get_cursor_position(self):
		return self.cursor
	
	def get_mouse_status(self):
		return self.mouse_pressed
	
	def get_events(self):
		events = []
		self.raw_keyups = []
		keyboard_only = True
		self.axes = [0.0, 0.0]
		pg_pressed = pygame.key.get_pressed()
		for event in pygame.event.get():
			if event.type in (pygame.MOUSEBUTTONDOWN, pygame.MOUSEBUTTONUP, pygame.MOUSEMOTION):
				move = event.type == pygame.MOUSEMOTION
				if move or event.button == 1:
					down = event.type == pygame.MOUSEBUTTONDOWN
					self.mouse_pressed = down
					x, y = event.pos
					x = x // 2
					y = y // 2
					self.mouse_events.append((x // 2, y // 2, move, down))
					self.cursor = (x, y)
			elif event.type == pygame.QUIT:
				self.quitAttempt = True
				return []
			elif event.type in (pygame.KEYDOWN, pygame.KEYUP):
				down = event.type == pygame.KEYDOWN
				if not down:
					self.raw_keyups.append(event.key)
				if down and event.key == pygame.K_F4:
					if pg_pressed[pygame.K_LALT] or pg_pressed[pygame.K_RALT]:
						self.quitAttempt = True
						return []
				elif down and event.key == pygame.K_ESCAPE:
					self.quitAttempt = True
					return []
				
				action = self._key_mapping.get(event.key, None)
				
				if action != None:
					self.my_pressed[action] = down
					events.append(MyEvent(action, down))
		self.axes[0] = 2.0 if self.my_pressed['right'] else 0.0
		self.axes[0] = -2.0 if self.my_pressed['left'] else self.axes[0]
		self.axes[1] = 2.0 if self.my_pressed['down'] else 0.0
		self.axes[1] = -2.0 if self.my_pressed['up'] else self.axes[1]

		joystick = None
		config = None
		any_axes_found = False
		if self.active_joystick != -1:
			config = self.joysticks[self.active_joystick]
			name = config.get('name', '')
			for js in self.actual_joysticks:
				name2 = js.get_name()
				if name.lower() == name2.lower():
					joystick = js
					break
		
		if joystick != None and config != None:
			cached_poll = {}
			for action in ('right', 'left', 'up', 'down', 'start', 'spray', 'walkie'):
				direction = False
				c = config.get(action, None)
				x = False
				if c != None:
					n = c[1]
					if c[0] == 'axis':
						direction = True
						x = cached_poll.get('a' + str(n))
						if x == None:
							x = joystick.get_axis(n)
							if abs(x) < 0.01:
								x = 0
							cached_poll['a' + str(n)] = x
						
						if c[2][1] == '+':
							if x < 0:
								x = 0
						else:
							if x > 0:
								x = 0
							else:
								x *= -1
						x = abs(x)
						
						if x > 0.01:
							keyboard_only = False
							if not any_axes_found:
								any_axes_found = True
								self.axes = [0.0, 0.0]
						
					elif c[0] == 'hat':
						direction = True
						x = cached_poll.get('h' + str(n))
						if x == None:
							x = joystick.get_hat(n)
							cached_poll['h' + str(n)] = x
						if c[2][0] == 'x':
							x = x[0]
						else:
							x = x[1]
						
						if c[2][1] == '+':
							if x < 0:
								x = 0
						else:
							if x > 0:
								x = 0
							else:
								x *= -1
						x = abs(x)
					elif c[0] == 'button':
						x = cached_poll.get('b' + str(n))
						if x == None:
							x = joystick.get_button(n)
							cached_poll['b' + str(n)] = x
						
					
					if action in ('start', 'spray', 'walkie'):
						pushed = x
						if direction:
							pushed = x >= .5
						if self.my_pressed[action] != pushed:
							self.my_pressed[action] = pushed
							events.append(MyEvent(action, pushed))
					elif action in ('left', 'right', 'down', 'up'):
						
						toggled = False
						if direction:
							pushed = x + 0.0
						else:
							pushed = 0.0 if (pushed < .5) else 1.0
						toggled = pushed > .2
						if pushed < 0.01:
							pushed = 0
						
						if toggled != self.my_pressed[action]:
							self.my_pressed[action] = toggled
							events.append(MyEvent(action, toggled))
						
						if action == 'left':
							if abs(self.axes[0]) < 0.01 and pushed > 0.01:
								self.axes[0] = -2 * pushed
						
						elif action == 'right':
							if abs(self.axes[0]) < 0.01 and pushed > 0.01:
								self.axes[0] = 2 * pushed
								
						elif action == 'up':
							if abs(self.axes[1]) < 0.01 and pushed > 0.01:
								self.axes[1] = -2 * pushed
								
						elif action == 'down':
							if abs(self.axes[1]) < 0.01 and pushed > 0.01:
								self.axes[1] = 2 * pushed
		
		self.axes[0] = self.axes[0] / 1.8
		
		x = self.axes[0]
		y = self.axes[1]
		
		if not any_axes_found and x != 0 and y != 0:
			xsign = 1 if (x > 0) else -1
			ysign = 1 if (y > 0) else -1
			x = 1.2 * xsign
			y = 1.2 * ysign
			
		if abs(x) < 0.05 and abs(y) < 0.05:
			rx = 0.0
			ry = 0.0
			
		else:
			ang = 3.14159265 / 4.0
			c = math.cos(ang)
			s = math.sin(ang)
			rx = x * c + y * s
			ry = -x * s + y * c
		
		self.axes[0] = rx / 1.8
		self.axes[1] = ry / 1.8
		
		return events
			
	def activate_joysticks(self):
		self.actual_joysticks = []
		active_joystick_name = None
		if self.active_joystick != -1:
			active_joystick_name = trim(self.joysticks[self.active_joystick].get('name', '')).lower()
		for i in range(pygame.joystick.get_count()):
			js = pygame.joystick.Joystick(i)
			js.init()
			self.actual_joysticks.append(js)
			name = trim(js.get_name()).lower()
			if name == active_joystick_name:
				self.active_actual_joystick = i
		
		if self.active_actual_joystick == -1:
			self.active_joystick = -1
			
	def verify_axis_value(self, x):
		return len(x) == 2 and x[0] in 'xy' and x[1] in '-+'
	
	def save_config(self):
		output = []
		if self.active_joystick != -1 and self.active_joystick < len(self.joysticks):
			output.append('#active: ' + self.joysticks[self.active_joystick].get('name', ''))
		else:
			output.append('#active: ')
		for joystick in self.joysticks:
			output.append(self._save_joystick(joystick))
		output = '$'.join(output)
		write_file('data/input_config.txt', output)
	
	
	def _save_joystick(self, config):
		output = []
		for key in config.keys():
			if key == 'name':
				value = config[key]
			else:
				value = ' '.join(safe_map(str, config[key]))
			
			row = '#' + key + ': ' + value
			
			output.append(row)
			
		return '\r\n'.join(output)
	
	def read_config_save(self):
		prev = read_file('data/input_config.txt')
		if prev != None:
			data = trim(prev).split('$')
			if len(data) > 0:
				active = trim(data[0])
				parts = active.split(':')
				active_joystick_name = None
				if len(parts) >= 2 and trim(parts[0]) == '#active':
					active_joystick_name = trim(':'.join(parts[1:]))
				for config in data[1:]:
					lines = config.split('\n')
					data = {}
					for line in lines:
						line = trim(line)
						if len(line) > 0 and line[0] == '#':
							parts = line[1:].split(':')
							if len(parts) == 2:
								key = trim(parts[0])
								value = trim(parts[1]).split(' ')
								if value[0] == 'axis':
									if len(value) == 3:
										n = parseInt(value[1])
										if self.verify_axis_value(value[2]):
											data[key] = ('axis', n, value[2])
								elif value[0] == 'button':
									if len(value) == 2:
										n = parseInt(value[1])
										data[key] = ('button', n)
								elif value[0] == 'hat':
									if len(value) == 3:
										n = parseInt(value[1])
										if self.verify_axis_value(value[2]):
											data[key] = ('hat', n, value[2])
								elif key == 'name':
									data[key] = trim(' '.join(value))
					name = data.get('name', None)
					if name != None and len(name) > 0:
						if active_joystick_name != None and active_joystick_name.lower() == name.lower():
							self.active_joystick = len(self.joysticks)
						self.joysticks.append(data)
	
	def get_mouse_events(self):
		output = self.mouse_events
		self.mouse_events = []
		return output

_jukebox = None

def get_jukebox():
	global _jukebox
	if _jukebox == None:
		_jukebox = JukeBox()
	return _jukebox

def play_sound(path):
	get_jukebox().play_sound(path)

class JukeBox:
	def __init__(self):
		pygame.mixer.init()
		self.current = None
		
		ps = get_persistent_state()
		if (ps.forever.get('sfx') != None):
			self.set_music_volume(ps.get_int_forever('music'))
			self.set_sfx_volume(ps.get_int_forever('sfx'))
		else:
			self.set_music_volume(70)
			self.set_sfx_volume(70)
		self.sounds = {}
		self.music_map = {
			'intro':'biologytake2',
			'12-0':'chemistry',
			'13-0':'chemistry',
			'14-0':'chemistry',
			'17-3':'chemistry',
			'18-0':'chemistry',
			'19a-0':'chemistry',
			'19b-1':'chemistry',
			'19-0':'chemistry',
			'99-0':'bossmusic',
		}
		
		
	def get_song_normalization(self, name):
		if name == 'title':
			return 1
		
		return .3
	
	def set_music_volume(self, percent):
		percent = max(0, min(100, int(percent)))
		self.music_volume = percent
		self.update_volume(self.current)
	
	def update_volume(self, song):
		if song != None:
			volume = self.music_volume * self.get_song_normalization(song)
			pygame.mixer.music.set_volume(volume / 100.0)
	
	def set_sfx_volume(self, percent):
		percent = max(0, min(100, int(percent)))
		self.sfx_volume = percent
		self.sounds = {}
	
	def get_music_volume(self):
		return self.music_volume
	
	def get_sfx_volume(self):
		return self.sfx_volume
	
	def play_sound(self, path):
		snd = self.sounds.get(path)
		if snd == None:
			fpath = 'sound/SFX/' + path + '.ogg'
			fpath = fpath.replace('.wav', '.ogg').replace('/', os.sep).replace('\\', os.sep)
			fpath = fpath.replace('.ogg.ogg', '.ogg')
			if not os.path.exists(fpath):
				fpath = fpath.replace('.ogg', '.wav')
			snd = self.sounds.get(fpath)
			if snd == None:
				snd = pygame.mixer.Sound(fpath)
				volume = self.sfx_volume / 100.0
				if path.startswith('talk'):
					if 'high' in path:
						volume = volume / 2
					volume = volume / 3
				if 'menumove' in path:
					volume = volume / 4
				snd.set_volume(volume)
				self.sounds[path] = snd
			else:
				self.sounds[path] = snd
		snd.play()
	
	def ensure_current_song(self, song):
		if song == 'bossmusic' and self.current == 'stringtheory':
			self.ensure_current_song('stringtheory')
			return
		if song == None:
			pygame.mixer.music.stop()
		else:
			song = 'sound/music/' + song + '.mp3'
			song = song.replace('/', os.sep).replace('\\', os.sep)
			if self.current != song:
				self.current = song
				self.update_volume(song)
				pygame.mixer.music.load(song)
				pygame.mixer.music.play(-1)
				pygame.mixer.music.play(-1)
	
	def get_song_for_level(self, level):
		return self.music_map.get(level, 'astrophysics')
	
	def update(self, levelname):
		song = self.get_song_for_level(levelname)
		#if levelname == '99-0' and self.current == 'stringtheory' and song != 'stringtheory':
			
		#	#return
			
		self.ensure_current_song(song)


def make_key(x, y, z):
	return str(str(x) + '|' + str(y) + '|' + str(z))

def make_ckey(c):
	return str(c[0]) + '|' + str(c[1]) + '|' + str(c[2])

class Circuits:
	def __init__(self, level):
		self.level = level
		self.refresh_groups(True)
		
	def refresh_charges(self):
		goo_was_fried = False
		on_circuits = get_hacks_for_level(self.level.name, 'on_circuits')
		if on_circuits == None:
			on_circuits = []
		active_input_panels = []
		for input_panel in self.power_in:
			x = input_panel[0]
			y = input_panel[1]
			z = input_panel[2]
			lookup = self.level.cellLookup[x][y]
			if z + 1 < len(lookup):
				ti = lookup[z + 1]
				if ti != None:
					t = self.level.grid[x][y][ti]
					if t != None and t.id == '45':
						active_input_panels.append(input_panel)
		on_groups = safe_map(lambda x: self.groups_by_coords[make_ckey(x)], on_circuits + active_input_panels + self.permanently_on)
		group_id = 1
		max_group_id = len(self.coords_by_group) - 1
		while group_id <= max_group_id:
			is_group_on = group_id in on_groups
			for coord in self.coords_by_group[group_id]:
				tile = self.level.get_tile_at(coord)
				if tile.actual_circuit:
					id = tile.id
					if id.endswith('on'):
						id = id[:-2]
					id = ((id + 'on') if is_group_on else id)
					tile = get_tile_store().get_tile(id)
					self.level.modify_block(coord[0], coord[1], coord[2], tile)
					if is_group_on:
						goo_check = self.level.get_tile_at(coord[0], coord[1], coord[2] + 1)
						if goo_check != None and goo_check.is_goo:
							self.level.modify_block(coord[0], coord[1], coord[2] + 1, None)
							goo_was_fried = True
			group_id += 1
		self.on_groups = on_groups
		if goo_was_fried:
			pass
			# TODO: play frying goo noise
		
	def refresh_groups(self, first_time=False):
		self.groups = []
		permanently_on = []
		grid = self.level.grid
		width = self.level.width
		height = self.level.height
		circuits = []
		circuits_by_loc = {}
		self.power_in = []
		y = 0
		while y < height:
			x = 0
			while x < width:
				stack = grid[x][y]
				z = 0
				for item in stack:
					if item == None:
						z += 1
					else:
						if item.circuit:
							circuits.append((x, y, z))
							circuits_by_loc[str(x)+'|' + str(y) + '|' + str(z)] = []
							if first_time and item.id.endswith('on'):
								permanently_on.append((x, y, z))
						if item.power_input:
							self.power_in.append((x, y, z))
						z += item.height
				x += 1
			y += 1
		
		tags = {} # key -> group name
		
		for circuit in circuits:
			x = circuit[0]
			y = circuit[1]
			z = circuit[2]
			neighbors = [
				circuits_by_loc.get(make_key(x, y - 1, z)),
				circuits_by_loc.get(make_key(x, y + 1, z)),
				circuits_by_loc.get(make_key(x + 1, y, z)),
				circuits_by_loc.get(make_key(x - 1, y, z))]
			for neighbor in neighbors:
				if neighbor != None:
					neighbor.append(circuit)
		
		group_id = 0
		while len(circuits) > 0:
			group_id += 1
			circuit = circuits.pop()
			while tags.get(make_ckey(circuit)) != None:
				if len(circuits) > 0:
					circuit = circuits.pop()
				else:
					circuit = None
					break
			if circuit != None:
				queue = [circuit]
				while len(queue) > 0:
					item = queue.pop()
					k = make_ckey(item)
					tags[k] = group_id
					for neighbor in circuits_by_loc[k]:
						if tags.get(make_ckey(neighbor)) == None:
							tags[make_ckey(neighbor)] = group_id
							queue.append(neighbor)
		
		groups_to_coords = [[]]
		for k in tags.keys():
			group_id = tags[k]
			while len(groups_to_coords) <= group_id:
				groups_to_coords.append([])
			coords = safe_map(int, k.split('|'))
			groups_to_coords[group_id].append((coords[0], coords[1], coords[2]))
		self.coords_by_group = groups_to_coords
		self.groups_by_coords = tags
		if first_time:
			self.permanently_on = permanently_on
		self.refresh_charges()
	
	def is_charged(self, x, y, z):
		k = make_key(x, y, z)
		group_id = self.groups_by_coords.get(k, 0)
		if group_id > 0:
			return group_id in self.on_groups
		return False
		

class CoilManager:
	
	def __init__(self, level):
		pass
	
	def set_fresh_coils(self, coils):
		self.coils = coils
	
	def find_zapped_sprites(self, sprites):
		coil_map = []
		for coil in self.coils:
			x = coil[0]
			while len(coil_map) <= x:
				coil_map.append([])
			coil_map[x].append(coil)
		
		for sprite in sprites:
			if sprite.death_counter < 0 and (sprite.main_or_hologram or sprite.israt):
				col = int(sprite.x // 16)
				while col + 2 >= len(coil_map):
					coil_map.append([])
				
				i = col - 1
				while i <= col + 1:
					coil_column = coil_map[i]
					for coil in coil_column:
						if abs(coil[2] - int(sprite.z // 8)) < 2:
							dx = coil[0] * 16 + 8 - sprite.x
							dy = coil[1] * 16 + 8 - sprite.y
							if dx * dx + dy * dy < 24 * 24:
								sprite.death_counter = 90
								sprite.death_type = 'bazat'
								sprite.immobilized = True
								
					i += 1
			

class HologramManager:
	def __init__(self, level):
		self.level = level
		pads = level.get_hologram_pads()
		self.copy_pad = None
		self.output_pads = []
		for pad in pads:
			if pad[3]:
				self.copy_pad = pad
			else:
				self.output_pads.append(pad)
		
		self.copy_mode_counter = -1
		self.used = False
		self.new_sprites = []
		self.sprites_created = []
		
	def get_new_sprites(self):
		ns = self.new_sprites
		self.new_sprites = []
		return ns
	
	def animation_sequence(self):
		if self.copy_mode_counter < 0:
			return None
		else:
			copy_process = self.copy_mode_counter > 0
			for sprite in self.sprites_created:
				sprite.clone_creating = copy_process
		return self.copy_mode_counter
	
	def update(self, playscene, level, player):
		if self.copy_pad == None: return
		
		self.copy_mode_counter -= 1
		if self.used:
			return
		
		cp = self.copy_pad
		
		col = int(player.x // 16)
		row = int(player.y // 16)
		layer = int(player.z // 8) - 1
		
		if cp[0] == col and cp[1] == row and cp[2] == layer:
			self.copy_mode_counter = 180
			# TODO: add new sprites
			self.used = True
			player.x = col * 16 + 8
			player.y = row * 16 + 8
			for p in self.output_pads:
				pc = p[0]
				pr = p[1]
				pl = p[2] + 1
				self.new_sprites.append(Sprite(pc * 16 + 8, pr * 16 + 8, pl * 8, 'hologram|main'))
				if level.name == '26-0' and playscene.story_mode:
					playscene.next = DialogScene(playscene, 'hologram')
			self.sprites_created = self.new_sprites[:]

class PushTracker:
	def __init__(self, sprite):
		self.push_counter = -1
		self.max_push_counter = 25
		self.push_target = None
		self.owner = sprite
	
class Level:
	def __init__(self, name, playscene):
		self.name = name
		self.initialize()
		self.sprite_z_sorter = lambda x,y:x.z < y.z
		self.push_trackers = []
		self.newsprites = []
		self.circuitry = Circuits(self)
		self.counter = 0
		self.render_exceptions = []
		self.complete = False
		self.spritify_counters = []
		self.moving_platforms = MovingPlatformManager(self)
		self.teleporters = TeleporterManager(self)
		self.switch_manager = SwitchManager(self, playscene)
		self.sprite_introducer = get_hacks_for_level(name, 'introduce_sprites')
		self.hologram_manager = HologramManager(self)
		self.coil_manager = CoilManager(self)
		self.counter = 0
		self.hologram_pads = []
	
	def get_push_tracker(self, sprite):
		for p in self.push_trackers:
			if p.owner == sprite:
				return p
		p = PushTracker(sprite)
		self.push_trackers.append(p)
		return p
	
	def get_hologram_pads(self):
		return self.hologram_pads
	
	def spray_from(self, sprites):
		success = False
		
		for sprite in sprites:
			col = int(sprite.x // 16)
			row = int(sprite.y // 16)
			layer = int(sprite.z // 8)
			dir = sprite.last_direction_of_movement
			x = sprite.x % 16
			y = sprite.y % 16
			
			if dir == 'n':
				if y < x:
					dir = 'ne'
				else:
					dir = 'nw'
			elif dir == 's':
				if y < x:
					dir = 'se'
				else:
					dir = 'sw'
			elif dir == 'e':
				if x + y > 16:
					dir = 'se'
				else:
					dir = 'ne'
			elif dir == 'w':
				if x + y > 16:
					dir = 'sw'
				else:
					dir = 'nw'
			
			
			if dir == 'nw':
				col -= 1
			elif dir == 'ne':
				row -= 1
			elif dir == 'sw':
				row += 1
			elif dir == 'se':
				col += 1
			
			maybe_goo = self.get_tile_at(col, row, layer)
			if maybe_goo != None and maybe_goo.is_goo:
				self.modify_block(col, row, layer, None)
				success = True
		return success
	
	def get_new_sprites(self):
		output = self.newsprites
		self.newsprites = []
		return output
	
	def initialize(self):
		self.rats = []
		lines = read_file('data/levels/' + self.name + '.txt').split('\n')
		values = {}
		for line in lines:
			line = trim(line)
			if len(line) > 0 and line[0] == '#':
				parts = line.split(':')
				if len(parts) > 1:
					key = parts[0][1:]
					value = ':'.join(parts[1:])
					values[key] = trim(value)
		self.values = values
		self.width = int(self.values['width'])
		self.height = int(self.values['height'])
		self.initialize_tiles(self.values['tiles'].split(','))
		
		self.initialize_switch_actions(values)
	
	def initialize_switch_actions(self, values):
		pswitches = {}
		nswitches = {}
		for key in values.keys():
			if key.startswith('action|'):
				sign = 1
				if key.startswith('action|negative|'):
					sign = -1
				name = key.split('|')[-1]
				value = values[key]
				if sign == 1:
					pswitches[name] = value
				else:
					nswitches[name] = value
				
		self.pswitches = pswitches
		self.nswitches = nswitches
	
	def activate_switch(self, name, is_positive):
		lookup = self.nswitches
		if is_positive:
			lookup = self.pswitches
		ts = get_tile_store()
		for change in lookup.get(name).split('%'):
			parts = change.split('^')
			col = int(parts[0])
			row = int(parts[1])
			layer = int(parts[2])
			id = parts[3]
			tile = None if (id == '0') else ts.get_tile(id)
			self.modify_block(col, row, layer, tile)
	
	def get_switches(self):
		return self.switches
	
	def get_coils(self):
		if self.coils == None:
			self.coils = []
			stacks = self.grid
			lookups = self.cellLookup
			y = 0
			while y < self.height:
				x = 0
				while x < self.width:
					z = 0
					stack = stacks[x][y]
					stacklen = len(stack)
					while z < stacklen:
						cell = stack[z]
						if cell != None and cell.teslaon:
							lookup = lookups[x][y]
							z = 0
							while z < len(lookup):
								t = lookup[z]
								if t != None and stack[t].teslaon:
									self.coils.append((x, y, z))
								z += 1
							break
						z += 1
					x += 1
				y += 1
		return self.coils
	
	def initialize_tiles(self, tiles):
		width = self.width
		height = self.height
		grid = make_grid(self.width, self.height, None)
		references = make_grid(self.width, self.height, None)
		moving_platforms = []
		self.switches = []
		self.coils = []
		self.hologram_pads = []
		self.teleporter_tiles = []
		tilestore = get_tile_store()
		i = 0
		for tile in tiles:
			x = i % width
			y = i // width
			tileStack = []
			referenceStack = []
			grid[x][y] = tileStack
			references[x][y] = referenceStack
			cells = tile.split('|')
			
			if len(cells) == 1 and len(cells[0]) == 0:
				pass
			else:
				for cell in cells:
					
					if cell == '0':
						referenceStack.append(None)
						tileStack.append(None)
					else:
						t = tilestore.get_tile(cell)
						if t.id == '16': # moving platform
							moving_platforms.append((x, y, len(referenceStack)))
						if t.id in ('t1', 't2', 't3'):
							self.teleporter_tiles.append((x, y, len(referenceStack), t.id))
						
						if t.isswitch:
							self.switches.append((x, y, len(referenceStack), t))
						
						if t.teslaon:
							self.coils.append((x, y, len(referenceStack)))
						
						if t.isholopad:
							self.hologram_pads.append((x, y, len(referenceStack), t.id == 'clone'))
						
						if t.id in ('rat', 'rat2', 'rat3', 'rat4'):
							if t.id == 'rat':
								dir = 'se'
							elif t.id == 'rat2':
								dir = 'ne'
							elif t.id == 'rat3':
								dir = 'sw'
							else:
								dir = 'nw'
							sprite = Sprite(x * 16 + 8, y * 16 + 8, len(tileStack) * 8, 'rat|' + dir)
							self.rats.append(sprite)
							sprite.standingon = tileStack[-1] # This will explode if a rat is placed on the bottom row of the map
							
							referenceStack.append(None)
							tileStack.append(None)
						else:
								
							z = 0
							while z < t.height:
								referenceStack.append(len(tileStack))
								z += 1
							
							tileStack.append(t)
				
			
			i += 1
		self.grid = grid
		self.cellLookup = references
		self.moving_platforms = moving_platforms
		
	def get_tile_at(self, x, y=None, z=None):
		if y == None:
			y = x[1]
			z = x[2]
			x = x[0]
		if z < 0: return None
		lookup = self.cellLookup[x][y]
		if len(lookup) <= z:
			return None
		index = lookup[z]
		if index == None:
			return None
		return self.grid[x][y][index]
	
	def get_platform_below(self, col, row, z, blocking_only):
	
		# TODO: interlace the block list here
		# so that blocks get included in the collision
		# results
		
		# Alternatively, pushed blocks can be part of the 
		# primary model and will be updated when they move
		# around.
		refStack = self.cellLookup[col][row]
		tileStack = self.grid[col][row]
		i = min(len(refStack) - 1, z // 8)
		
		while i >= 0:
			t = refStack[i]
			if t == None:
				pass
			else:
				tile = self.grid[col][row][t]
				if not blocking_only or tile.blocking:
					return ((i + 1) * 8, tile)
			i -= 1
	
	def render(self, screen, xOffset, yOffset, sprites, render_counter):
		self.counter += 1
		self.allsprites = sprites
		self.any_res = False
		self.re_blits = {}
		
		width = self.width
		height = self.height
		sprite_lookup = {}
		for sprite in sprites:
			x = int(sprite.x // 16)
			y = int(sprite.y // 16)
			key = str(x) + '_' + str(y)
			list = sprite_lookup.get(key)
			if list == None:
				list = []
				sprite_lookup[key] = list
			list.append(sprite)
		empty_list = []
		
		re_on_keys = {}
		re_off_keys = {}
		for re in self.render_exceptions:
			re_on_keys[re.on_key] = re_on_keys.get(re.on_key, [])
			re_on_keys[re.on_key].append(re)
			re_off_keys[re.off_key] = re_off_keys.get(re.off_key, [])
			re_off_keys[re.off_key].append(re)
		
		z_sorter = lambda a,b: a.z < b.z
		for k in re_on_keys.keys():
			re_on_keys[k] = safe_sorted(re_on_keys[k], z_sorter)
		for k in re_off_keys.keys():
			re_off_keys[k] = safe_sorted(re_off_keys[k], z_sorter)
		
		i = 0
		while i < width + height:
			col = i
			row = 0
			
			while row < height and col >= 0:
				if col < width:
					sprite_list = sprite_lookup.get(str(col) + '_' + str(row))
					k = str(col) + '|' + str(row)
					re_on = re_on_keys.get(k)
					re_off = re_off_keys.get(k)
					
					if re_on != None:
						re_on = re_on[:]
					if re_off != None:
						re_off = re_off[:]
					
					self.render_tile_stack(screen, col, row, xOffset, yOffset, render_counter, sprite_list, re_on, re_off)
				row += 1
				col -= 1
			i += 1
		
		
	def render_sprite(self, screen, sprite, xOffset, yOffset, render_counter, re_render=False):
		
		q = sprite.render_me(screen, xOffset, yOffset, render_counter)
		if not re_render:
			if sprite.x % 16 < 6 and sprite.y % 16 > 10:
				if sprite.main_or_hologram or sprite.israt:
					col = int(sprite.x // 16)
					row = int(sprite.y // 16)
					layer = int(sprite.z // 8)
					k = str(col) + '^' + str(row)
					r = (layer + sprite.height, q)
					if self.re_blits.get(k) == None:
						self.re_blits[k] = [r]
					else:
						if self.re_blits[k][-1][0] < r[0]:
							self.re_blits[k].append(r)
						else:
							i = 0
							while i < len(self.re_blits[k]):
								if self.re_blits[k][i][0] > r[0]:
									self.re_blits[k] = self.re_blits[k][:i] + [r] + self.re_blits[k][i:]
									break
								i += 1
		
	def render_tile_stack(self, screen, col, row, xOffset, yOffset, render_counter, sprites, re_on, re_off):
		re_sprite_on = None
		re_sprite_off = None
		re_block_on = None
		re_block_off = None
		
		re_blits = self.re_blits.get(str(col + 1) + '^' + str(row - 1))
		
		if re_on != None:
			re_block_on = []
			re_sprite_on = []
			for re in re_on:
				if re.is_block:
					self.any_res = True
					re_block_on.append(re)
				else:
					re_sprite_on.append(re)
			
		if re_off != None:
			re_block_off = []
			re_sprite_off = []
			for re in re_off:
				if re.is_block:
					self.any_res = True
					re_block_off.append(re)
				else:
					re_sprite_off.append(re)				
		
		prefix = str(col) + '|' + str(row) + '|'
		
		stack = self.grid[col][row]
			
		z = 0
		x = xOffset + col * 16 - row * 16 - 16
		y = yOffset + col * 8 + row * 8
		i = 0
		while i < len(stack):
			
			# re blit stuff from the right
			
			if re_blits != None:
				while re_blits != None and len(re_blits) > 0 and re_blits[0][0] <= z:
					q = re_blits[0][1]
					sprite = q[0]
					sprite.render_me(screen, q[1], q[2], q[3])
					re_blits = re_blits[1:]
			
			tile = stack[i]
			if sprites != None:
				new_sprites = []
				for sprite in sprites:
					if sprite.z <= z * 8:
						if re_sprite_off != None and len(re_sprite_off) > 0 and re_sprite_off[0].z == z and re_sprite_off[0].tile == sprite:
							re = re_sprite_off.pop(0)
						else:
							re_offset = (0, 0)
							if re_sprite_on != None:
								rei = 0
								while rei < len(re_sprite_on):
									re = re_sprite_on[rei]
									if re.tile == sprite:
										re_sprite_on.pop(rei)
										re_offset = re.get_offset()
									else:
										rei += 1
							
							self.render_sprite(screen, sprite, xOffset + re_offset[0], yOffset + re_offset[1], render_counter)
					else:
						new_sprites.append(sprite)
				sprites = new_sprites
			
			while re_sprite_on != None and len(re_sprite_on) > 0 and re_sprite_on[0].z < z:
				re = re_sprite_on.pop(0)
				offset = re.get_offset()
				_x = xOffset + offset[0]
				_y = yOffset + offset[1]
				xdiff = re.do_show[0] - re.dont_show[0]
				ydiff = re.do_show[1] - re.dont_show[1]
				xdiff_pixel = xdiff * 16 - ydiff * 16
				ydiff_pixel = xdiff * 8 + ydiff * 8
				self.render_sprite(screen, re.tile, _x + xdiff_pixel, _y + ydiff_pixel, render_counter)
				
			if tile == None:
				if re_block_on != None and len(re_block_on) > 0 and re_block_on[0].z == z:
					re = re_block_on.pop(0)
					offset = re.get_offset()
					re.tile.render(screen, x + offset[0], y - z * 8 + offset[1], render_counter)
					#z += re.tile.height - 1
				z += 1
			else:
				if re_block_off != None and len(re_block_off) > 0 and re_block_off[0].z == z:
					re = re_block_off.pop(0)
				elif re_block_on != None and len(re_block_on) > 0 and re_block_on[0].z == z:
					re = re_block_on.pop(0)
					offset = re.get_offset()
					tile.render(screen, x + offset[0], y - z * 8 + offset[1], render_counter)
				else:
					tile.render(screen, x, y - z * 8, render_counter)
				z += tile.height
			i += 1
		
		
		if re_blits != None:
			
			while re_blits != None and len(re_blits) > 0 and re_blits[0][0] <= z:
				q = re_blits[0][1]
				sprite = q[0]
				sprite.render_me(screen, q[1], q[2], q[3])
				re_blits = re_blits[1:]
		
		
		if (sprites != None and len(sprites) > 0) or (re_sprite_on != None and len(re_sprite_on) > 0):
			sprites = [] if sprites == None else safe_sorted(sprites, self.sprite_z_sorter)
			sprite_i = 0
			re_i = 0
			if re_sprite_on == None:
				re_sprite_on = []
			
			rendered_sprites = []
			for re in re_sprite_on:
				for sprite in sprites:
					if re.tile == sprite:
						rendered_sprites.append(sprite)
			if len(rendered_sprites) > 0:
				new_sprites = []
				for sprite in rendered_sprites:
					if sprite in rendered_sprites:
						pass
					else:
						new_sprites.append(sprite)
				sprites = new_sprites
			
			while sprite_i < len(sprites) or re_i < len(re_sprite_on):
				sprite = None
				re = None
				if sprite_i == len(sprites):
					re = re_sprite_on[re_i]
					re_i += 1
				elif re_i == len(re_sprite_on):
					sprite = sprites[sprite_i]
					sprite_i += 1
				else:
					re = re_sprite_on[re_i]
					sprite = sprites[sprite_i]
					if re.z < int(sprite.z // 8):
						sprite = None
						re_i += 1
					else:
						re = None
						sprite_i += 1
				
				if re != None:
					offset = re.get_offset()
					_x = xOffset + offset[0]
					_y = yOffset + offset[1]
					if re.direction == 'NE' or re.direction == 'NW':
						xdiff = re.do_show[0] - re.dont_show[0]
						ydiff = re.do_show[1] - re.dont_show[1]
						xdiff_pixel = xdiff * 16 - ydiff * 16
						ydiff_pixel = xdiff * 8 + ydiff * 8
						_x += xdiff_pixel
						_y += ydiff_pixel
					
					self.render_sprite(screen, re.tile, _x, _y, render_counter)

				else:
					suppress = False
					if re_sprite_off != None:
						rei = 0
						while rei < len(re_sprite_off):
							re = re_sprite_off[rei]
							if sprite == re.tile:
								re_sprite_off.pop(rei)
								suppress = True
							else:
								rei += 1
					if not suppress:
						self.render_sprite(screen, sprite, xOffset, yOffset, render_counter)
	
	# there are no blockages. It's already been verified by the time this function
	# has been called.
	def push_block(self, sprite, start_col, start_row, end_col, end_row, layer):
		play_sound('blockpush.wav')
		pt = self.get_push_tracker(sprite)
		pt.push_counter = -1
		pt.push_target = None
		
		block = self.modify_block(start_col, start_row, layer, None)
		self.modify_block(end_col, end_row, layer, block)
		
		if start_row == end_row:
			if start_col < end_col:
				dir = 'SE'
			else:
				dir = 'NW'
		elif start_row < end_row:
			dir = 'SW'
		else:
			dir = 'NE'
			
		self.render_exceptions.append(RenderException((start_col, start_row, layer), dir, block, True))
		
		below_layer = layer - 1
		
		was_standing_on = self.get_tile_at(start_col, start_row, below_layer)
		
		target_lookup = self.cellLookup[end_col][end_row]
		stack = self.grid[end_col][end_row]
		should_spritify = False
		if below_layer < len(target_lookup):
			standingon = target_lookup[below_layer]
			if standingon == None:
				should_spritify = True
			else:
				standingon = stack[standingon]
				if not standingon.blocking:
					should_spritify = True
				else:
					if standingon.id == 'po' and block.id == '46' and self.circuitry.is_charged(end_col, end_row, below_layer):
						play_sound('battery_charge.wav')
						self.modify_block(end_col, end_row, layer, get_tile_store().get_tile('45'))
					if standingon.id == 'pi':
						if block.id == '45':
							play_sound('electricity_flows.wav')
							self.circuitry.refresh_charges()
					if was_standing_on != None and was_standing_on.id == 'pi' and block.id == '45':
						play_sound('battery_deplete.wav')
						self.modify_block(end_col, end_row, layer, get_tile_store().get_tile('46'))
						self.circuitry.refresh_charges()
						
		else:
			should_spritify = True
		
		if should_spritify:
			
			self.spritify_counters.append([end_col, end_row, layer, 5])
			#self.spritify_block(end_col, end_row, layer)
	
	def spritify_block(self, col, row, layer):
		block = self.modify_block(col, row, layer, None)
		self.newsprites.append(Sprite(col * 16 + 8, row * 16 + 8, layer * 8, 'block|' + block.id))
		
	
	def modify_block(self, col, row, layer, type):
		self.coils = None
		output = None
		z = 0
		stack = self.grid[col][row]
		newstack = []
		layer_bottom = layer
		layer_top = layer
		if type != None:
			layer_top += type.height - 1
		
		for item in stack:
			if item == None:
				newstack.append(None)
				z += 1
			else:
				
				if z >= layer_bottom and z <= layer_top:
					if z == layer_bottom:
						output = item
					newstack += [None] * item.height
				else:
					newstack.append(item)
				z += item.height
		while z <= layer_top:
			newstack.append(None)
			z += 1
		# at this point newstack has all None's where the modification will go
		
		i = 0
		z = 0
		while i < len(newstack):
			item = newstack[i]
			if z == layer_bottom:
				newstack[i] = type
				for q in range(layer_top - layer_bottom):
					newstack.pop(i + 1)
				break
			if item == None:
				z += 1
			else:
				z += item.height
			i += 1
		
		teleported = False
		if type != None and type.pushable:
			# check to see if we just placed a pushable block onto a teleporter platform
			t = self.get_tile_at(col, row, layer - 1)
			if t != None and t.teleporter:
				d = self.teleporters.get_destination(col, row, layer - 1)
				if d == 'blocked':
					self.announce_teleporter_blocked()
				elif d == None:
					pass
				else:
					teleported = True
					source = (col, row, layer)
					self.teleporters.teleport_block(type, source, d)
		
		if not teleported:
			copy_array(stack, newstack)
			self.canonicalize_stack(col, row)
		
		return output
	
	def announce_teleporter_blocked(self):
		pass
		# TODO: play sound
	
	def canonicalize_stack(self, col, row):
		stack = self.grid[col][row]
		lookup = []
		i = 0
		for item in stack:
			if item == None:
				lookup.append(None)
			else:
				lookup += [i] * item.height
			i += 1
		copy_array(self.cellLookup[col][row], lookup)
	
	def update(self, sprites, sprite_additions, sprite_removals):
		created = None
		if self.sprite_introducer != None:
			created = self.sprite_introducer(self, self.counter)
		
		if created != None:
			for c in created:
				col = int(c.x // 16)
				row = int(c.y // 16)
				layer = int(c.z // 8)
				c.standingon = self.get_tile_at(col, row, layer - 1)
				sprite_additions.append(c)
		
		new_re = []
		self.moving_platforms.update(sprites, new_re)
		self.teleporters.update()
		
		for sprite in self.teleporters.get_new_sprites():
			sprite_additions.append(sprite)
		for sprite in self.teleporters.get_removed_sprites():
			sprite_removals.append(sprite)
		
		
		for re in self.render_exceptions:
			re.update()
			if not re.expired:
				new_re.append(re)
		self.render_exceptions = new_re
		
		i = 0
		while i < len(self.spritify_counters):
			t = self.spritify_counters[i]
			if t[-1] == 0:
				self.spritify_counters = self.spritify_counters[:i] + self.spritify_counters[i + 1:]
				self.spritify_block(t[0], t[1], t[2])
			else:
				t[-1] -= 1
				i += 1
		
		
		self.counter += 1
		
	
	def get_moving_platforms(self):
		return self.moving_platforms
	
	def get_teleporters(self):
		return self.teleporter_tiles
	

_level_manager = [None, None]

def is_story_mode():
	quiet_mode = 1 == get_persisted_session_int('quiet_mode')
	return not quiet_mode

def get_level_manager():
	global _level_manager
	story_mode = is_story_mode()
	i = 1 if story_mode else 0
	if _level_manager[i] == None:
		_level_manager[i] = LevelManager(story_mode)
	return _level_manager[i]

class LevelManager:
	def __init__(self, story_mode):
		self.ordering = [
				('1-3', "Lab Entrance", (4, 9, 1, 'ne')),
				('2-3', "Fill 'er up!", (7, 13, 2, 'ne')),
				('2a-0', "Around the Room", (6, 10, 1, 'ne')),
				('3-1', "The Green Room", (9, 10, 1, 'ne')),
				('4-0', "Got the Blues", (5, 8, 1, 'ne')),
				('5-0', "One Block, Two Block", (2, 10, 1, 'ne')),
				('6-0', "Roadblock", (1, 9, 3, 'ne')),
				('7-0', "An Impasse", (1, 11, 3, 'ne')),
				('8-0', "Plug it in", (6, 9, 1, 'ne')),
				('9-0', "Energize", (1, 7, 1, 'ne')),
				('10-2', "You do this IRL", (5, 11, 1, 'ne')),
				('11-0', "Boom?", (1, 4, 3, 'se')),
				('12-0', "A Block Too Far", (1, 10, 3, 'ne')),
				('13-0', "Beam me up, Steve!", (7, 18, 1, 'ne')),
				('14-0', "Blocks in High Places", (5, 10, 3, 'ne')),
				('14a-0', "Jump the River", (4, 10, 2, 'ne')),
				('15-0', "Don't Look Down", (1, 10, 2, 'ne')),
				('16-0', "Islands", (10, 16, 2, 'ne')),
				('17-3', "Block Party", (9, 18, 3, 'ne')),
				('18-0', "Power Overwhelming", (1, 19, 2, 'se')),
				('19a-0', "Like Men", (2, 7, 1, 'se')),
				('19b-1', "Rats in a Maze", (1, 17, 1, 'se')),
				('19c-0', "Co-operation", (2, 24, 3, 'ne')),
				('19-0', "A Shockin Experience", (1, 8, 1, 'se')),
				('20-0', "1.21 \"Jiga\" Watts", (1, 15, 2, 'se')),
				('21-0', "R.O.U.S.", (9, 13, 2, 'ne')),
				('21a-0', "High Voltage", (1, 11, 1, 'se')),
				('flipmaze', "On The Flip Side", (1, 19, 1, 'se')),
				('25-0', "Hunger Game", (1, 12, 1, 'se')),
				('24-0', "Three Blind Mice", (1, 20, 3, 'ne')),
				('26-0', "Mirror Mirror", (1, 9, 1, 'se')),
				('27-0', "Many Me", (1, 9, 1, 'se')),
				('28-0', "Simon Says", (1, 12, 1, 'se')),
				('90-0', "Here There Be Ratmen", (1, 9, 3, 'se')),
				('99-0', "Doctor What?", (8, 4, 1, 'nw'))]
		if story_mode:
			self.ordering = [('intro', "Your Cubicle", (12, 14, 1, 'nw'))] + self.ordering
	
	def get_current_level_index(self, current_level_name):
		for i in range(len(self.ordering)):
			if self.ordering[i][0] == current_level_name:
				return i
		return None
	
	def get_current_room_name(self, current_level_name):
		return self.ordering[self.get_current_level_index(current_level_name)][1]
	
	def get_next_level(self, current_level_name):
		i = self.get_current_level_index(current_level_name)
		next_i = i + 1
		if next_i < len(self.ordering):
			return self.ordering[next_i][0]
		return None
	
	def get_starting_point_for_level(self, name):
		i = self.get_current_level_index(name)
		if len(self.ordering[i]) >= 3:
			return self.ordering[i][2]
		return None
	
	def get_starting_level(self):
		return self.ordering[0]
	


"""

on_circuits:
	pick 1 circuit PER CIRCUIT GROUP to indicate that that group should
	always remain on.
	A power input or output pad counts as a circuit so something like...
		---wire---[pad]---wire---
	...is 1 group, not 2.
	
	Format is a list of tuples that are (x, y, and z) coordinates. 
	You can get these from the map editor title bar

moving_platforms:
	directions map to platforms in tile parser order.
	lowest y first, then lowest x, then lowest z
	
	P is pause
"""

def make_sprite(type, col, row, layer):
	return Sprite(col * 16 + 8, row * 16 + 8, layer * 8, type)

def _hack_do_stuff_99(level, counter):
	pass #if counter == 1:
		

def _hack_introduce_sprites_intro(level, counter):
	
	if counter == 120:
		s = make_sprite('supervisor', 14, 0, 1)
		s.set_automation(Automation(level, 'intro_supervisor'))
		return [s]
	elif counter == 150:
		s = make_sprite('janitor', 14, 0, 1)
		s.set_automation(Automation(level, 'intro_janitor'))
		return [s]
	return []

def _hack_do_stuff_intro(playscene, level, counter):
	if counter == -1:
		playscene.player.set_automation(Automation(level, 'intro_protagonist'))
	
	if counter == 329 + 14:
		playscene.next = DialogScene(playscene, 'intro')
	
	if counter == 1000:
		playscene.next = TransitionScene(playscene, PlayScene(get_level_manager().get_next_level('intro'), True))


_level_specific_hacks = {
	'intro': {
		'introduce_sprites': _hack_introduce_sprites_intro,
		'do_stuff': _hack_do_stuff_intro,
		'dialog': { } #'transition_level_one': _hack_dialog_intro_transition_level_one }
	},
	
	'99-0': {
		
	},
	
	'9-0': {
		'on_circuits': [
			(1, 2, 0)
		]
	},
	
	'15-0': {
		'moving_platforms': [
			'NW P P SE SE SE P P NW NW', 
			'SW P P NE NE NE P P SW SW'
		]
	},
	
	'16-0': {
		'moving_platforms': [
			'SW SW SW SW SW SW SW P P NE NE NE NE NE NE NE P P',
			'NE NE NE NE P P SW SW SW SW P P',
			'NW NW NW NW P P SE SE SE SE P P'
		]
	},
	
	'17-3': {
		'moving_platforms': [
			'SW SW SW SW SW P P NE NE NE NE NE P P'
		]
	},
	
	'18-0': {
		'moving_platforms': [
			'SE SE NW NW NW NW NW NW NW NW NW NW SE SE SE SE SE SE SE SE',
			'NW NW SE SE SE SE SE SE SE SE SE SE NW NW NW NW NW NW NW NW',
			'NW NW NW NW NW NW NW NW NW SE SE SE SE SE SE SE SE SE SE NW',
			'NW NW NW NW SE SE SE SE SE SE SE SE SE SE NW NW NW NW NW NW',
			'NW NW NW NW NW NW NW SE SE SE SE SE SE SE SE SE SE NW NW NW',
			'NW SE SE SE SE SE SE SE SE SE SE NW NW NW NW NW NW NW NW NW',
			'SE SE SE SE NW NW NW NW NW NW NW NW NW NW SE SE SE SE SE SE',
			'NW NW NW NW NW NW NW SE SE SE SE SE SE SE SE SE SE NW NW NW',
			'SE SE SE SE SE SE SE SE SE SE NW NW NW NW NW NW NW NW NW NW',
			'NW NW SE SE SE SE SE SE SE SE SE SE NW NW NW NW NW NW NW NW',
			'NW NW NW NW NW NW NW NW NW NW SE SE SE SE SE SE SE SE SE SE'
		]
	},
	
	'19c-0': {
		'moving_platforms': [
			'SW SW SW SW SW P P NE NE NE NE NE P P'
		]
	},
	
	'21-0': {
		'moving_platforms': [
			'NW NW P P SE SE SE SE P P NW NW'
		]
	},
	
	'25-0': {
		'moving_platforms': [
			'SE SE NW NW NW NW NW SE SE SE',
			'SE SE SE SE NW NW NW NW NW SE',
			'NW NW NW NW NW SE SE SE SE SE',
			'SE SE NW NW NW NW NW SE SE SE',
			'SE SE SE SE SE NW NW NW NW NW',
			'SE NW NW NW NW NW SE SE SE SE',
			'SE SE SE SE NW NW NW NW NW SE',
			'SE SE SE NW NW NW NW NW SE SE',
			'NW NW NW NW NW SE SE SE SE SE'
		]
	},
	
	'24-0': {
		'moving_platforms': [
			'P P SE SE SE SE SE SE P P NW NW NW NW NW NW',
			'NE P P SW SW P P NE',
			'NE NE NE NE NE NE NE NE P P SW SW SW SW SW SW SW SW P P',
			'P SW SW P P NE NE P'
		]
	},
	
	'90-0': {
		'moving_platforms': [
			'SW SW SW P P NE NE NE P P'
		]
	}
}

def get_hacks_for_level(name, category):
	global _level_specific_hacks
	dict = _level_specific_hacks.get(name)
	if dict != None:
		return dict.get(category)
	return None

class MovingPlatformManager:
	
	def __init__(self, level):
		self.level = level
		self.ticker = 0
		self.pause_tokens = {}
		self.initialize_pause_tokens(level.name)
		
		directions = get_hacks_for_level(level.name, 'moving_platforms')
		if directions == None:
			directions = []
		else:
			directions = safe_map(lambda x:x.split(), directions)
		self.platforms = self.level.get_moving_platforms()
		self.num_platforms = len(self.platforms)
		while len(directions) < self.num_platforms:
			directions.append(['P'])
		self.directions = directions
	
	def set_pause_token(self, i, on):
		self.pause_tokens[str(i)] = on
	
	def initialize_pause_tokens(self, name):
		pt = self.pause_tokens
		if name == '16-0':
			pt['0'] = True
		elif name == '18-0':
			for i in range(0, 11):
				pt[str(i)] = True
		elif name == '24-0':
			pt['2'] = True
		elif name == '25-0':
			for i in range(9):
				pt[str(i)] = True
		elif name == '90-0':
			pt['0'] = True
	
	def is_paused_platform(self, i):
		if self.pause_tokens.get(str(i)) == True:
			return True
		return False
	
	def update(self, sprites, render_exceptions):
		self.ticker += 1
		if self.ticker % 60 == 0:
			level = self.level
			i = 0
			while i < self.num_platforms:
				
				if not self.is_paused_platform(i):
				
					platform = self.platforms[i]
					directions = self.directions[i]
					direction = directions[0]
					if direction == 'P':
						directions.pop(0)
						directions.append('P')
					else:
						target = [platform[0], platform[1], platform[2]]
						
						if direction == 'NW':
							target[0] -= 1
							offset = (-1, 0)
						elif direction == 'NE':
							target[1] -= 1
							offset = (0, -1)
						elif direction == 'SW':
							target[1] += 1
							offset = (0, 1)
						else:
							target[0] += 1
							offset = (1, 0)
						t_lower = level.get_tile_at(target[0], target[1], target[2])
						t_upper = level.get_tile_at(target[0], target[1], target[2] + 1)
						if t_lower == None and t_upper == None:
							move_us = [platform]
							
							z = move_us[-1][2] + 2
							col = platform[0]
							row = platform[1]
							lookupStack = level.cellLookup[col][row]
							stack = level.grid[col][row]
							spritification = []
							while z < len(lookupStack):
								if lookupStack[z] != None:
									tile = stack[lookupStack[z]]
									if tile != None:
										if tile.pushable:
											
											neighbor_lookup = level.cellLookup[col + offset[0]][row + offset[1]]
											neighbor_stack = level.grid[col + offset[0]][row + offset[1]]
											blocked = False
											if len(spritification) == 0:
												for z_offset in (z + 0, z + 1):
													if len(neighbor_lookup) <= z_offset:
														pass
													else:
														
														if z_offset < 0:
															ntile = None
														else:
															ntile = neighbor_lookup[z_offset]
														if ntile != None:
															ntile = neighbor_stack[ntile]
															if ntile != None and ntile.blocking:
																blocked = True
																break
											else:
												blocked = True
											loc = [col, row, z]
											if blocked:
												spritification.append(loc)
											else:
												move_us.append(loc)
												
											z += tile.height - 1
										else:
											break
									else:
										break
								else:
									break
								z += 1
							
							for block in spritification:
								level.spritify_block(block[0], block[1], block[2])
							
							first = True
							for move_me in move_us:
								mp = level.modify_block(move_me[0], move_me[1], move_me[2], None)
								target = [move_me[0] + offset[0], move_me[1] + offset[1], move_me[2]]
								level.modify_block(target[0], target[1], target[2], mp)
								
								if first:
									self.platforms[i] = target
								render_exceptions.append(RenderException(move_me, direction, mp, True))
								first = False
							directions.append(directions.pop(0))
							
							
							for sprite in sprites:
								if sprite.standingon != None and int(sprite.x // 16) == move_us[-1][0] and int(sprite.y // 16) == move_us[-1][1] and int(sprite.z // 8) == move_us[-1][2] + 2:
									sx = int(sprite.x // 16)
									if sx == move_us[-1][0]:
										sy = int(sprite.y // 16)
										sz = int(sprite.z // 8)
										if sy == move_us[-1][1] and sz == move_us[-1][2] + 2:
											coords = (int(sprite.x // 16), int(sprite.y // 16), int(sprite.z // 8))
											target = [coords[0] + offset[0], coords[1] + offset[1], coords[2]]
											lookup = level.cellLookup[target[0]][target[1]]
											stack = level.grid[target[0]][target[1]]
											blocked = False
											for _i in range(sprite.height):
												check = [target[0], target[1], target[2] + _i]
												t = level.get_tile_at(check)
												if t != None and t.blocking:
													blocked = True
											if not blocked:
												render_exceptions.append(RenderException(coords, direction, sprite, False))
												sprite.x += offset[0] * 16
												sprite.y += offset[1] * 16
											else:
												sprite.standingon = None
							
				i += 1



_render_exceptions_coords = [
(2, 1), # 7
(4, 2), # 6
(6, 3), # 5
(8, 4), # 4
(10, 5), # 3
(12, 6), # 2
(14, 7), # 1
]

class RenderException:
	
	# start is a set of 3D coordinates
	def __init__(self, start, direction, tile, is_block):
		global _render_exceptions_coords
		self.is_block = is_block
		self.is_sprite = not is_block
		self.z = start[2]
		self.counter = 0
		self.expired = False
		if direction == 'NE' or direction == 'NW':
			self.do_show = start
			if direction == 'NE':
				self.dont_show = (start[0], start[1] - 1, start[2])
			else: #NW
				self.dont_show = (start[0] - 1, start[1], start[2])
		else: # SW || SE
			self.dont_show = start
			if direction == 'SE':
				self.do_show = (start[0] + 1, start[1], start[2])
			else: #SW
				self.do_show = (start[0], start[1] + 1, start[2])
		self.direction = direction
		self.tile = tile
		self.on_key = str(self.do_show[0]) + '|' + str(self.do_show[1])
		self.off_key = str(self.dont_show[0]) + '|' + str(self.dont_show[1])
		
		if self.direction == 'NW':
			flipper = lambda x:(-x[0], -x[1])
			reverse = 1
		elif self.direction == 'NE':
			flipper = lambda x:(x[0], -x[1])
			reverse = 1
		elif self.direction == 'SW':
			flipper = lambda x:(x[0], -x[1])
			reverse = -1
		elif self.direction == 'SE':
			flipper = lambda x:(-x[0], -x[1])
			reverse = -1
			
		self.offsets = safe_map(flipper, _render_exceptions_coords[::reverse])
	
	def update(self):
		self.counter += 1
		if self.counter > 6:
			self.expired = True
	
	def get_offset(self):
		return self.offsets[min(6, self.counter)]
		



_switch_mapping = {
	'4-0': [
		'blue'
	],
	
	'5-0': [
		'bridge',
		None,
		None
	],
	
	'8-0': [
		'exit'
	],
	
	'9-0': [
		'exit'
	],
	
	'10-2': [
		'bridgene',
		'bridgenw',
		'bridgese'
	],
	
	'12-0': [
		'exit'
	],
	
	'14-0': [
		'blue',
		'exit'
	],
	
	'16-0': [
		None,
		'green'
	],
	
	'17-3': [
		'red',
		'blue'
	],
	
	'18-0': [None] * 9,
	
	'19-0': [
		'power'
	],
	
	'19c-0':
	[ 'button' ],
	
	'20-0': [
		'blue',
		'power'
	],
	
	'21-0': [
		'green',
		'blue',
		'powere',
		'powerw'
	],
	
	'21a-0': [
		'green',
		'magenta',
		'yellow',
		'cyan'
	],
	
	'24-0': [
		'green',
		'power',
		'purple',
		'grey',
		None,
		'yellow'
	],
	
	'25-0': [
		'grey'
	],
	
	'26-0': [
		'blue',
		'red'
	],
	
	'27-0': [
		'cyan',
		'magenta',
		'yellow'
	],
	
	'28-0': [
		'door',
		'bridge'
	],
	
	'90-0': [
		None,
		'green',
		'blue',
		'power',
		'power',
		'yellow'
	]
}

def override_switch_behavior(manager, level, index, is_blue):
	enabled = manager.enabled
	name = level.name
	
	if name == '5-0':
		if index > 0:
			if enabled[1] and enabled[2]:
				level.activate_switch('door', True)
			else:
				level.activate_switch('door', False)
				
			return True
	
	elif name == '16-0':
		if index == 0:
			# Don't confuse the meaning of these 0's.
			# The index == 0 which is the switch ID
			# The pause token is 0 which is the moving platform ID
			# They're the same here because that's a coincidence
			level.moving_platforms.set_pause_token('0', not enabled[index])
			return True
	
	elif name == '18-0':
		mapping = [
			[10],
			[0],
			[2],
			[4],
			[1, 9],
			[5],
			[7],
			[8],
			[3, 6]
		]
		for platform in mapping[index]:
			level.moving_platforms.set_pause_token(str(platform), not enabled[index])
		return True
	
	
	elif name == 'flipmaze':
		level.activate_switch('blue', is_blue)
		return True
		
	elif name == '24-0':
		if index == 4:
			level.moving_platforms.set_pause_token('2', not enabled[index])
			return True
	
	elif name == '25-0':
		level.moving_platforms.set_pause_token(str(index - 1), not enabled[index])
		if index != 0:
		
			return True
	elif name == '90-0':
		if index == 0:
			level.moving_platforms.set_pause_token('0', not enabled[index])
			return True
	return False


class SwitchManager:
	def __init__(self, level, playscene):
		self.playscene = playscene
		self.level = level
		self.flipmaze_hack = playscene.story_mode and level.name == 'flipmaze'
		self.switches = self.level.get_switches()
		self.enabled = [False] * len(self.switches)
		self.statuses = [None] * len(self.switches)
		self.locations = {}
		for switch in self.switches:
			k = str(switch[0]) + '^' + str(switch[1])
			self.locations[k] = self.locations.get(k, [])
			self.locations[k].append(switch)
		
		ts = get_tile_store()
		self.rubiks = ts.get_tile('rubiks')
		self.colors = {
			# Switch, block
			'gray': (ts.get_tile('b1'), ts.get_tile('3')),
			'red': (ts.get_tile('b2'), ts.get_tile('9')),
			'blue': (ts.get_tile('b4'), ts.get_tile('10')),
			'green': (ts.get_tile('b3'), ts.get_tile('11')),
			'magenta': (ts.get_tile('b6'), ts.get_tile('12')),
			'cyan': (ts.get_tile('b5'), ts.get_tile('13')),
			'yellow': (ts.get_tile('b7'), ts.get_tile('14')),
			'battery': (ts.get_tile('pi'), ts.get_tile('45')),
		}
		
		self.activator_lookup = {}
		for color in self.colors.keys():
			self.activator_lookup[self.colors[color][0].id] = self.colors[color][1]
		
	
	def update_statuses_for_sprite(self, sprite, level):
		col = int(sprite.x // 16)
		row = int(sprite.y // 16)
		layer = int(sprite.z // 8) - 1
		
		floor = level.get_tile_at(col, row, layer)
		if floor != None and floor.isswitch:
			i = 0
			while i < len(self.switches):
				sw = self.switches[i]
				
				if sw[0] == col and sw[1] == row and sw[2] == layer:
					self.statuses[i] = ('sprite', sprite)
					break
				i += 1
	
	def check_switch_for_block_and_update(self, i, switch, level):
		col = switch[0]
		row = switch[1]
		layer = switch[2] + 1
		above = level.get_tile_at(col, row, layer)
		if above != None and above.pushable:
			self.statuses[i] = ('block', above)
		
		
	def update_statuses(self, sprites):
		if len(self.switches) == 0:
			return
		
		i = 0
		while i < len(self.switches):
			self.statuses[i] = None
			i += 1
		
		players = []
		level = self.level
		for sprite in sprites:
			if sprite.main_or_hologram:
				players.append(sprite)
			else:
				self.update_statuses_for_sprite(sprite, level)
		for p in players:
			self.update_statuses_for_sprite(p, level)
		
		i = 0
		while i < len(self.switches):
			self.check_switch_for_block_and_update(i, self.switches[i], level)
			i += 1
		
	def update_enabled(self, sprites, suppress_triggers):
		
		self.update_statuses(sprites)
		before = self.enabled[:]
		
		i = 0
		while i < len(self.switches):
			switch = self.switches[i]
			status = self.statuses[i]
			if status != None:
				if status[0] == 'sprite':
					if switch[3] == self.colors['gray'][0] or self.level.name == 'flipmaze':
						if self.flipmaze_hack:
							self.flipmaze_hack = False
							self.playscene.next = DialogScene(self.playscene, 'h1')
						self.enabled[i] = True
				elif status[0] == 'block':
					type = switch[3]
					if type == self.colors['gray'] or (status[1] == self.rubiks and type.id != 'pi'):
						self.enabled[i] = True
					elif self.activator_lookup[type.id] == status[1]:
						self.enabled[i] = True
					else:
						self.enabled[i] = False
			else:
				self.enabled[i] = False
			
			if not suppress_triggers:
				if before[i] != self.enabled[i]:
					self.do_action(self.level, self.level.name, i, self.enabled[i])
					
			
			i += 1
		
	def do_action(self, level, name, switch_index, positive):
		global _switch_mapping
		
		is_blue = level.get_tile_at(self.switches[switch_index]) == self.colors['blue'][0]
		
		if override_switch_behavior(self, level, switch_index, is_blue):
			return
		
		mapping = _switch_mapping.get(name)
		if mapping != None:
			action_name = mapping[switch_index]
			level.activate_switch(action_name, positive)
		#print "Level", name, switch_index, positive
		#pass
		

class TeleporterManager:
	def __init__(self, level):
		self.level = level
		transporters = level.get_teleporters()
		self.counter = 0
		senders = []
		receivers = []
		enabled = []
		in_use = []
		for t in transporters:
			if t[3] == 't1' or t[3] == 't3':
				senders.append(t)
				enabled.append(t[3] == 't1')
			else:
				receivers.append(t)
		
		self.senders = senders
		self.receivers = receivers
		
		self.manual_level_config(level.name)

		self.new_sprites = {}
		self.remove_sprites = []
		self.in_use = len(self.senders) * [999]
		self.enabled = enabled

	
	def manual_level_config(self, name):
		if name == '13-0':
			r = self.receivers
			self.receivers = [
				r[0], r[9], r[4], r[5], r[2], r[10], r[3], r[1], r[8], r[6], r[7]
			]
	
	def set_teleporter_status(self, col, row, layer, isenabled):
		i = 0
		total = len(self.senders)
		while i < total:
			sender = self.senders[i]
			if sender[0] == col and sender[1] == row and sender[2] == layer:
				enabled[i] = isenabled
				self.update_tiles(i)
				return
			receiver = self.receivers[i]
			if receiver[0] == col and receiver[1] == row and receiver[2] == layer:
				enabled[i] = isenabled
				self.update_tiles(i)
				return
	
	def enable_teleporter(self, col, row, layer):
		self.set_teleporter_status(col, row, layer, True)
	
	def disable_teleporter(self, col, row, layer):
		self.set_teleporter_status(col, row, layer, False)
	
	def update_tiles(self, index):
		sender = self.senders[index]
		receiver = self.receivers[index]
		level = self.level
		ts = get_tile_store()
		s_tile = ts.get_tile('t1')
		r_tile = ts.get_tile('t2')
		if not self.enabled[index]:
			s_tile = ts.get_tile('t3')
			r_tile = ts.get_tile('t3')
		level.modify_block(sender[0], sender[1], sender[2], s_tile)
		level.modify_block(receiver[0], receiver[1], receiver[2], r_tile)
	
	def get_destination(self, col, row, layer):
		final = self.get_sender((col, row, layer))
		if final != None and self.enabled[final]:
			output = self.receivers[final]
			for i in range(4):
				z = output[2] + i + 1
				tile = self.level.get_tile_at(output[0], output[1], z)
				if tile != None and tile.blocking:
					return 'blocked'
			if self.in_use[final] < 200:
				return None
			self.in_use[final] = 0
			return output
		return None

	def teleport_block(self, blocktype, source, target):
		
		sprite = Sprite(source[0] * 16 + 8, source[1] * 16 + 8, source[2] * 8, 'block|'+blocktype.id)
		self.teleport_sprite(sprite, target)
	
	def teleport_sprite(self, sprite, target):
		play_sound('teleport.wav')
		sprite.immobilized = True
		self.remove_sprites.append(sprite)
		
		self.new_sprites[self.counter + 1] = self.new_sprites.get(self.counter + 1, [])
		self.new_sprites[self.counter + 1].append(Sprite(sprite.x, sprite.y, sprite.z, 'teleport|' + sprite.type))
		
		self.new_sprites[self.counter + 120] = self.new_sprites.get(self.counter + 120, [])
		z = (target[2] + 1) * 8
		clone = Sprite(target[0] * 16 + 8, target[1] * 16 + 8, z, 'receiving|' + sprite.type)
		clone.standingon = get_tile_store().get_tile('t2')
		clone.prototype = sprite
		
		self.new_sprites[self.counter + 120].append(clone)
	
	def tag_as_in_use(self, coords):
		sender_index = self.get_sender(coords)
		if sender_index != None:
			self.in_use[sender_index] = 0
			
	def is_in_use(self, i):
		return self.in_use[i] <= 90
	
	def get_sender(self, coords):
		i = 0
		while i < len(self.senders):
			s = self.senders[i]
			if s[0] == coords[0] and s[1] == coords[1] and s[2] == coords[2]:
				return i
			i += 1
		return None
	
	def get_new_sprites(self):
		new_sprites = self.new_sprites.get(self.counter, None)
		if new_sprites == None:
			return []
		return self.new_sprites.pop(self.counter)
		
	
	def get_removed_sprites(self):
		output = self.remove_sprites
		self.remove_sprites = []
		return output
	
	def update(self):
		self.counter += 1
		i = 0
		while i < len(self.in_use):
			self.in_use[i] += 1
			i += 1
	

def flags_contains(flags, items):
	for c in items:
		if c in flags:
			return True
	return False

class Tile:
	def is_num(self, string):
		return string in '0123456789'
		
	def __init__(self, id, images, height, flags):
		self.id = id
		images = images.split('|')
		self.framerate = 4
		if self.is_num(images[0][0]):
			self.framerate = int(images[0])
			images = images[1:]
		self.hascharge = False
		
		if id == '4off':
			images = ['teslacoil1.png']
		
		self.tesla = id in ('4off', '4')
		self.teslaon = id == '4'
		
		self.images = safe_map(lambda x:get_image('tiles/' + x), images)
		self.still = len(self.images) == 1
		self.still_image = self.images[0]
		self.y_offsets = safe_map(lambda x:24 - x.get_height(), self.images)
		self.still_y_offset = self.y_offsets[0]
		self.image_count = len(self.images)
		self.height = height
		self.no_blocks = False
		self.isholopad = id in ('clone', 'clone2')
		self.pushable = flags_contains(flags, 's')
		self.blocking = flags_contains(flags, 'x')
		self.circuit = flags_contains(flags, 'eh')
		self.actual_circuit = flags_contains(flags, 'e')
		self.stairs = flags_contains(flags, '12345678')
		self.teleporter = id == 't1'
		self.power_input = id == 'pi'
		self.power_output = id == 'po'
		self.isexit = id in ('37', '38')
		self.cant_push_over = flags_contains(flags, 'n')
		self.research = id == '41'
		self.goo = id in ('42', '43', '44')
		self.isswitch = id in ('pi', 'b1', 'b2', 'b3', 'b4', 'b5', 'b6', 'b7')
		self.powerup = self.research or self.goo
		self.is_goo = id == '15'
		self.goo_size = 0
		if self.goo:
			self.goo_size = ((int(id) - 41) - 1) * 2 + 1 # 1, 3, 5 OMGHAX
		self.blocking = self.blocking or self.stairs or self.pushable
		if self.stairs:
			self.no_blocks = True
			double = flags_contains(flags, '5678')
			topo = 2 if double else 1
			if flags_contains(flags, '15'):
				self.topography = [0, -topo, -topo, 0]
				self.entrance = 'SE'
			elif flags_contains(flags, '26'):
				self.topography = [0, 0, -topo, -topo]
				self.entrance = 'SW'
			elif flags_contains(flags, '37'):
				self.topography = [-topo, 0, 0, -topo]
				self.entrance = 'NW'
			elif flags_contains(flags, '48'):
				self.topography = [-topo, -topo, 0, 0]
				self.entrance = 'NE'
			
	# Code duplicated below
	def get_image(self, render_counter):
		if self.still:
			return self.still_image
		i = (render_counter // self.framerate) % self.image_count
		return self.images[i]

	# Code duplicated above
	def render(self, screen, x, y, render_counter):
		
		if self.still:
			img = self.still_image
			y += self.still_y_offset
		else:
			i = (render_counter // self.framerate) % self.image_count
			img = self.images[i]
			y += self.y_offsets[i]
		
		screen.blit(img, (x, y))
		

class TileStore:
	def __init__(self):
		tiles = {}
		manifest_dir = canonicalize_path('data/tile_manifests')
		manifests = os.listdir(manifest_dir)
		for manifest in manifests:
			if '.svn' in manifest: continue
			manifest_data = read_file(manifest_dir + os.sep + manifest).split('\n')
			for line in manifest_data:
				if line != '' and line[0] != '#':
					cols = line.split('\t')
					if len(cols) >= 4:
						id = cols[0]
						images = cols[1]
						height = int(cols[2])
						flags = cols[3]
						tiles[id] = Tile(id, images, height, flags)
		self.tiles = tiles
	
	def get_tile(self, id):
		return self.tiles[id]
					
	def get_all_block_tiles(self):
		output = []
		for k in self.tiles.keys():
			if self.tiles[k].pushable:
				output.append(k)
		return output


_tile_store = None
def get_tile_store():
	global _tile_store
	if _tile_store == None:
		_tile_store = TileStore()
	return _tile_store

# persistent state cannot save strings with newlines across "forever" sessions

_persistent_state = None

def get_persistent_state():
	global _persistent_state
	if _persistent_state == None:
		_persistent_state = PersistentState()
	return _persistent_state

def load_persistent_state():
	get_persistent_state().load_game()

def save_persistent_state():
	get_persistent_state().save_game()

def get_persisted_forever_int(name):
	return get_persistent_state().get_int_forever(name)
def get_persisted_session_int(name):
	return get_persistent_state().get_int_session(name)
def get_persisted_level_int(name):
	return get_persistent_state().get_int_level(name)

def get_persisted_forever_string(name):
	return get_persistent_state().get_string_forever(name)
def get_persisted_session_string(name):
	return get_persistent_state().get_string_session(name)
def get_persisted_level_string(name):
	return get_persistent_state().get_string_level(name)

def set_persisted_forever_int(name, value):
	return get_persistent_state().set_int_forever(name, value)
def set_persisted_session_int(name, value):
	return get_persistent_state().set_int_session(name, value)
def set_persisted_level_int(name, value):
	return get_persistent_state().set_int_level(name, value)

def set_persisted_forever_string(name, value):
	return get_persistent_state().set_string_forever(name, value)
def set_persisted_session_string(name, value):
	return get_persistent_state().set_string_session(name, value)
def set_persisted_level_string(name, value):
	return get_persistent_state().set_string_level(name, value)

def increment_persisted_forever_int(name, amount):
	set_persisted_forever_int(name, get_persisted_forever_int(name) + amount)

def increment_persisted_level_int(name, amount):
	set_persisted_level_int(name, get_persisted_level_int(name) + amount)

def increment_persisted_session_int(name, amount):
	set_persisted_session_int(name, get_persisted_session_int(name) + amount)

def persistence_change_level():
	get_persistent_state().level = {}

class PersistentState:
	def __init__(self):
		self.forever = {}
		self.session = {}
		self.level = {}
	
	def purge_all(self):
		self.forever = {}
		self.session = {}
		self.level = {}
	
	def save_game(self):
		write_file('data/save.txt', self.serialize())
	
	def change_level(self):
		self.level = {}
	
	def serialize(self):
		output = []
		for key in self.forever.keys():
			value = self.forever[key]
			if value[1] == None:
				value[1] = ''
			output.append(value[0] + key + ':' + str(value[1]))
		return '\n'.join(output)
	
	def load_game(self):
		t = read_file('data/save.txt')
		values = {}
		if t != None:
			lines = t.split('\n')
			for line in lines:
				parts = line.split(':')
				if len(parts) >= 2:
					key = parts[0]
					value = ':'.join(parts[1:])
					if len(key) >= 2:
						type = key[0]
						name = key[1:]
						if type == 'i':
							try:
								value = int(value)
							except:
								value = 0
							
						else:
							type = 's'
						values[name] = (type, value)
		self.forever = values
						
					
	
	def set_int_forever(self, name, value):
		self._set_int(name, value, self.forever)
	
	def set_int_session(self, name, value):
		self._set_int(name, value, self.session)
	
	def set_int_level(self, name, value):
		self._set_int(name, value, self.level)
	
	def get_int_forever(self, name):
		return self._get_int(name, self.forever)
	
	def get_int_session(self, name):
		return self._get_int(name, self.session)
	
	def get_int_level(self, name):
		return self._get_int(name, self.level)
	
	def set_string_forever(self, name, value):
		self._set_string(name, value, self.forever)
	
	def set_string_session(self, name, value):
		self._set_string(name, value, self.session)
	
	def set_string_level(self, name, value):
		self._set_string(name, value, self.level)
	
	def get_string_forever(self, name):
		return self._get_string(name, self.forever)
	
	def get_string_session(self, name):
		return self._get_string(name, self.session)
	
	def get_string_level(self, name):
		return self._get_string(name, self.level)
	
	def _set_int(self, name, value, lookup):
		lookup[name] = ('i', value)
	
	def _set_string(self, name, value, lookup):
		lookup[name] = ('s', value)
	
	def _get_int(self, name, lookup):
		value = lookup.get(name)
		if value != None and value[0] == 'i':
			return value[1]
		return 0
	
	def _get_string(self, name, lookup):
		value = lookup.get(name)
		if value != None and value[0] == 's':
			return value[1]
		return ''

class ClickyButton:
	def __init__(self, id, x, y, text, on_color, off_color, pressed_color, action, get_label):
		self.id = id
		self.text = text
		self.on_color = on_color
		self.off_color = off_color
		self.pressed_color = pressed_color
		self.pressed = False
		self.on_color = on_color
		self.off_color = off_color
		self.pressed_color = pressed_color
		self.left = x
		self.top = y
		size = 14
		self.size = size
		self.enabled = True
		if text != None:
			self.on_img = get_text(text, size, on_color)
			self.off_img = get_text(text, size, off_color)
			self.pressed_img = get_text(text, size, pressed_color)
			self.right = x + self.off_img.get_width()
			self.bottom = y + self.off_img.get_height()
		else:
			self.right = x + 10
			self.bottom = y + get_text("!y", size, (0, 0, 0)).get_height()
		self.im = get_input_manager()
		self.action = action
		self.get_label = get_label
		self.last_text = None
	
	def is_mouse_over(self):
		x, y = self.im.get_cursor_position()
		if self.left > x or self.right < x or self.top > y or self.bottom < y:
			return False
		return True
	
	def is_pressing(self):
		return self.is_mouse_over() and self.im.get_mouse_status()
		
	def render(self, screen):
		if self.get_label != None:
			label = self.get_label(self.id)
			if label == '---': # horrible hack
				self.enabled = False
			else:
				self.enabled = True
			if label != self.last_text:
				self.pressed_img = get_text(label, self.size, self.pressed_color)
				self.on_img = get_text(label, self.size, self.on_color)
				self.off_img = get_text(label, self.size, self.off_color)
				self.bottom = self.on_img.get_height() + self.top
				self.right = self.on_img.get_width() + self.left
		
		if not self.enabled:
			img = self.off_img
		elif self.is_pressing():
			img = self.pressed_img
		elif self.is_mouse_over():
			img = self.on_img
		else:
			img = self.off_img
		
		screen.blit(img, (self.left, self.top))
	
	def on_click(self):
		if self.enabled:
			self.action(self.id)

class ConfigureInputScene:
	def __init__(self):
		self.next = self
		hb = self.handle_button
		a = (255, 255, 0)
		b = (255, 255, 255)
		c = (128, 128, 255)
		self.labels = []
		heading = (200, 200, 200)
		heading_size = 20
		command_color = (128, 180, 255)
		key_command_color = (255, 255, 255)
		use_mouse_label_size = 18
		
		js_option_y_coords = []
		self.key_command_color = key_command_color
		use_mouse_label = get_text("Use the mouse for this screen", use_mouse_label_size, (100, 255, 100))
		self.use_mouse_label = use_mouse_label
		
		js_label = get_text("Select Active Joystick", heading_size, heading)
		self.labels.append(((10, 10), js_label))
		y = 10 + 10 + js_label.get_height()
		x = 20
		js_option_y_coords.append(y)
		
		self.buttons = [ClickyButton('no joystick', x, y, "None", a, b, c, hb, None)]
		self.buttons.append(ClickyButton('exit', 300, 5, "Return to Main Menu", a, b, c, hb, None))
		im = get_input_manager()
		i = 1
		self.im = im
		for js in im.actual_joysticks:
			y += 20
			js_option_y_coords.append(y)
			self.buttons.append(ClickyButton('joystick ' + str(i), x, y, js.get_name(), a, b, c, hb, None))
			i += 1
		
		self.js_option_coords = js_option_y_coords
		
		y += 30
		gl = self.get_button_label
		self.labels.append(((10, y), get_text("Commands", heading_size, heading)))
		y += 20
		col1 = 20
		col2 = 115
		col3 = 205
		self.js_config_x = col3
		self.js_config_y = []
		self.keys = [
			None,
			('left', "Left", "Left Arrow"),
			('right', "Right", "Right Arrow"),
			('up', "Up", "Up Arrow"),
			('down', "Down", "Down Arrow"),
			('start', "Pause/Confirm", "Enter"),
			('spray', "Decontaminant", "Space"),
			('walkie', "Walkie Talkie", "W"),
			(None, "Quit", 'Esc / Alt+F4')]
		for command in self.keys:
			
			start_y = y
			self.js_config_y.append(y)
			if command == None:
				self.labels.append(((col2, y), get_text("Key", 14, heading)))
				self.labels.append(((col3, y), get_text("Joystick", 14, heading)))
			else:
				self.labels.append(((col1, y), get_text(command[1], 14, command_color)))
				if command[0] == None:
					self.labels.append(((col2, y), get_text(command[2], 14, key_command_color)))
				else:
					self.buttons.append(ClickyButton('keyconfig ' + command[0], col2, y, None, a, b, c, hb, gl))
					self.buttons.append(ClickyButton('jsconfig ' + command[0], col3, y, None, a, b, c, hb, gl))
			y += 16
	
	def get_button_label(self, id):
		parts = id.split(' ')
		if parts[0] == 'jsconfig':
			return self.im.get_config_label_for_key_for_active(parts[1])
		elif parts[0] == 'keyconfig':
			return self.im.get_config_label_for_key_for_keyboard(parts[1])
				
		return " "
	
	def save_changes(self):
		im = get_input_manager()
		im.save_config()
		im.save_key_config()
	
	def handle_button(self, id):
		im = get_input_manager()
		if id == 'exit':
			self.save_changes()
			self.next = TransitionScene(self, MainMenuScene())
		elif id == 'no joystick':
			im.set_active_actual_joystick(-1)
		else:
			parts = id.split(' ')
			if parts[0] == 'joystick':
				im.set_active_actual_joystick(int(parts[1]) - 1)
			elif parts[0] == 'keyconfig':
				action = parts[1]
				self.next = SetInputScene(action, True, self)
			elif parts[0] == 'jsconfig':
				action = parts[1]
				self.next = SetInputScene(action, False, self)
			else:
				self.next = TransitionScene(MainMenuScene(), self)
	
	def process_input(self, events, pressed, axes, mouse):
		for mouse_event in mouse:
			if mouse_event[2] == False and mouse_event[3] == False:
				for button in self.buttons:
					if button.is_mouse_over():
						button.on_click()
						break
		
		
	def update(self, counter):
		pass
	
	def render(self, screen, counter):
		im = get_input_manager()
		
		for button in self.buttons:
			button.render(screen)
		for label in self.labels:
			
		
			pos = label[0]
			img = label[1]
			screen.blit(img, pos)
		
		if ((counter // 25) % 2) == 1:
			screen.blit(self.use_mouse_label, (
				200 - self.use_mouse_label.get_width() // 2,
				300 - self.use_mouse_label.get_height()))
		
		if im.active_actual_joystick == -1:
			i = 0
		else:
			i = im.active_actual_joystick + 1
		x = 5
		y = self.js_option_coords[i]
		pygame.draw.rect(screen, (255, 0, 0), pygame.Rect(x, y + 2, 4, 4))

class CreditsScene:
	def __init__(self, canSkip):
		self.next = self
		self.canSkip = canSkip
		self.speed = .7
		self.counter = 0
		self.things = [
		[None, (' ', "Credits"), None],
		[None, ("Programming", "Blake O'Hare"), 'blake'],
		['spears', ("Level Art", "Angel McLaughlin"), None],
		[None, ("Music and Sound", "Adrian Cline"), 'ikanreed'],
		['satyrane', ("Story and Dialog", "Laura Freer"), None],
		[None, ("Portrait and Large Art", "\"Fixception\""), 'fixception'],
		['brendan', ("Puzzle Design", "Brendan & Steve"), 'stiva'],
		['duke', ("Level Implementation", "Will Duke & Ted Burton"), 'eofpi'],
		[None, (" ", "Thank you for playing."), None]
		]
		
		if not self.canSkip:
			r = get_persisted_forever_int('research')
			r += get_persisted_session_int('research')
			self.things.append(
			[None, (" ", "Research papers collected: " + str(r) + " out of 35"), None]
			)
		
		i = 0
		while i < len(self.things):
			thing = self.things[i]
			left = thing[0]
			right = thing[2]
			if left != None:
				left = get_image('us/' + left + '.png')
			if right != None:
				right = get_image('us/' + right + '.png')
				right = pygame.transform.flip(right, True, False)
			self.things[i][0] = left
			self.things[i][2] = right
			
			i += 1
		
	def process_input(self, events, pressed, axes, mouse):
		for event in events:
			if event.down:
				if event.key == 'start' or event.key == 'spray' or event.key == 'walkie':
					if self.canskip:
						self.next = TransitionScene(self, MainMenuScene())
	
	def update(self, counter):
		get_jukebox().ensure_current_song('stringtheory')
		self.counter += 1
	
	def render(self, screen, counter):
		
		y = 350 - int(self.counter * self.speed)# * 1.4)
		
		for item in self.things:
			left = item[0]
			right = item[2]
			top = get_text(item[1][0], 14, (123, 123, 123))
			bottom = get_text(item[1][1], 24, (255, 255, 255))
			
			x1 = screen.get_width() // 2 - top.get_width() // 2
			x2 = screen.get_width() // 2 - bottom.get_width() // 2
			x = min(x1, x2)
			screen.blit(top, (x1, y))
			y2 = y + top.get_height() + 5
			screen.blit(bottom, (x2, y2))
			r = 400 - x
			if left != None:
				screen.blit(left, (x - left.get_width() - 14, y - 5))
			if right != None:
				screen.blit(right, (r + 14, y - 5))
			y = y2 + bottom.get_height() + 140
		
		if y < -50:
			self.next = TransitionScene(self, MainMenuScene())

def get_startup_dialog(play_scene):
	id = {
		'1-3': 'a1',
		'2-3': 'a2',
		'4-0': 'b1',
		#'5-0': 'b2',
		#'6-0': 'b3',
		#'7-0': 'b4',
		'8-0': 'c1',
		'9-0': 'c2',
		#'10-2': 'c3',
		#'11-0': 'c4',
		'12-0': 'd1',
		#'13-0': 'd2',
		#'14-0': 'd3',
		'15-0': 'e1',
		#'16-0': 'e2',
		#'17-3': 'e3',
		'18-0': 'e4',
		'19a-0': 'f1',
		'19b-1': 'f2',
		'19-0': 'g1',
		'20-0': 'g2',
		#'21-0': 'g3',
		#'flipmaze': 'h1',
		'25-0': 'h2',
		#'24-0': 'i1',
		#'26-0': 'j1',
		'27-0': 'j2',
		#'28-0': 'j3',
		'90-0': '90',
		'99-0': '99'
	}.get(play_scene.level.name, None)
	
	if id != None:
		return DialogScene(play_scene, id)
	return None

class TextPrinter:
	def __init__(self, text, speaker):
		self.text = text
		self.counter = 0
		self.speaker= speaker
		
	def get_next_char(self, pressed):
		self.counter += 1
		t = ''
		if len(self.text) == 0:
			return None
		amt = 20 if (pressed and self.counter > 6) else 1
		t = self.text[:amt]
		self.text = self.text[amt:]
		if self.counter % 6 == 1:

			sound = None
			if self.speaker == 's':
				sound = 'talkbloop'
			elif self.speaker in 'jf':
				sound = 'talkblooplow'
			elif self.speaker == 'p':
				sound = 'talkbloophigh'
			
			if sound != None:
				play_sound(sound)
		return t


class DialogPause:
	def __init__(self, amount):
		self.remaining = amount
	
	def keep_pausing(self):
		self.remaining -= 1
		return self.remaining >= 0
		
	

class DialogScene:
	def __init__(self, playscene, id):
		self.id = id
		self.next = self
		self.wait_counter = 240 if (id == 'hologram') else 0
		self.phase_duration = 60
		self.playscene = playscene
		self.initialize_script(id)
		self.text_printer = None
		self.prompt_for_continue = False
		self.pauser = None
		self.surfaces = {}
		self.i = 0
		self.phase = 'init'
		self.phase_counter = 0
		self.active_portrait = None
		self.buffer = []
		self.clear_on_continue = False
		self.fast_speed = False
		self.colors = {
			'w':(255, 255, 255),
			's':(220, 180, 140),
			'j':(140, 180, 230),
			'f':(0, 200, 100),
			'p':(255, 120, 180)
		}
		self.buffer_clear_counter = -42
		self.frame_yielding = -42
		self.buffer_shift_offset = 0
	
	def initialize_script(self, id):
		file = read_file('data/dialog/' + id + '.txt')
		self.lines = file.split('\n')
		self.labels = {}
		i = 0
		while i < len(self.lines):
			parts = self.lines[i].split('|')
			if parts[0] == 'label' and len(parts) == 2:
				self.labels[parts[1]] = i
			i += 1
		
	def process_input(self, events, pressed, axes, mouse):
		pushed = False
		self.fast_speed = False
		for event in events:
			if event.key in ('start', 'spray', 'walkie') and event.down:
				pushed = True
				#self.fast_speed = True
		
		for j in ('start', 'spray', 'walkie'):
			if pressed[j]:
				self.fast_speed = True
		
		if pushed and self.prompt_for_continue:
			# TODO: this can be improved so it isn't quite as jumpy
			# should also flush all text_printer queues
			self.prompt_for_continue = False
			if self.clear_on_continue:
				self.buffer = []
				
		
	
	def update(self, counter):
		if self.wait_counter > 0:
			self.playscene.update(counter)
			self.wait_counter -= 1
			return
		self.phase_counter += 1
		self.frame_yielding -= 1
		
		if self.frame_yielding >= 0:
			for s in self.playscene.sprites:
				if s.issupervisor:
					#print 'supervisor:', s.automation.counter
					pass
			self.playscene.update(counter)
		
		keep_going = True
		while keep_going:
			if self.phase == 'init':
				if self.phase_counter == self.phase_duration:
					self.phase = 'dialog'
				keep_going = False
			elif self.phase == 'end':
				if self.phase_counter == self.phase_duration:
					self.next = self.playscene
					self.next.next = self.next
				keep_going = False
			elif self.prompt_for_continue:
				keep_going = False
			elif self.text_printer != None:
				n = self.text_printer.get_next_char(self.fast_speed)
				if n == None:
					self.text_printer = None
				else:
					self.buffer[-1][0] += n
					keep_going = False
			elif self.buffer_clear_counter > 0:
				self.buffer_clear_counter -= 1
				if self.buffer_clear_counter == 0:
					self.buffer = self.buffer[1:]
				keep_going = False
			elif self.pauser != None:
				if not self.pauser.keep_pausing():
					self.pauser = None
				keep_going = False
			else:
				if self.i < len(self.lines):
					line = trim(self.lines[self.i])
					if len(line) > 0:
						parts = line.split('|')
						command = parts[0]
						if command == 'i':
							self.change_portrait(parts[1])
						elif command == 'r':
							self.change_portrait(None)
						elif command == 'p':
							self.pauser = DialogPause(int(parts[1]))
						elif command == 's':
							if len(self.buffer) == 3:
								self.buffer_clear_counter = 10
								keep_going = False
								self.i -= 1
							else:
								color = self.colors[parts[1]]
								text = parts[2]
								self.buffer.append(['', color])
								self.text_printer = TextPrinter(text, parts[1])
						elif command == 'credits':
							self.next = TransitionScene(self, CreditsScene(False))
							keep_going = False
						elif command == 'music':
							get_jukebox().ensure_current_song(parts[1])
						elif command == 'snd':
							if not parts[1].startswith('talk'):
								play_sound(parts[1])
						elif command == 'c':
							self.prompt_for_continue = True
							self.clear_on_continue = False
							keep_going = False
						elif command == 'cc':
							self.prompt_for_continue = True
							self.clear_on_continue = True
						elif command == 'h':
							hack_function = get_hacks_for_level(self.playscene.level.name, 'dialog_hack')[parts[1]]
							hack_function(self.playscene, self.playscene.level)
						elif command == 'y':
							self.frame_yielding = int(parts[1])
						elif command == 'goto':
							self.do_goto(parts[1])
						elif command == 'if':
							var = parts[1]
							operator = parts[2]
							value = int(parts[3])
							label = parts[4]
							self.do_if(operator, var, value, label)
						elif command == 'end':
							self.phase_counter = 0
							self.phase = 'end'
					self.i += 1
				else:
					self.phase = 'end'
					self.phase_counter = 0
					keep_going = False
	
	def change_portrait(self, id):
		if id == None:
			self.active_portrait = None
		else:
			self.active_portrait = get_image('portraits/' + id + '.png')
		
	
	def do_goto(self, label):
		line = self.labels.get(label)
		if line != None:
			i = line
	
	def do_if(self, op, var, val, lbl):
		
		finders = [get_persisted_level_int, get_persisted_session_int, get_persisted_forever_int]
		value = 0
		for finder in finders:
			value = finder(var)
			if value != 0:
				break
		
		if op == '=':
			if value == val:
				self.do_goto(lbl)
		elif op == '<':
			if value < val:
				self.do_goto(lbl)
		elif op == '>':
			if value < val:
				self.do_goto(lbl)
		elif op == '~':
			if value != val:
				self.do_goto(lbl)
		
	
	def render_box(self, screen, phase, counter):
		
		width = 300
		height = 100
		if phase == 'end':
			counter = self.phase_duration - counter
		
		if phase == 'dialog':
			pass
		else:
			half = self.phase_duration // 2
			if counter < self.phase_duration // 2:
				height = 10
			else:
				height = 90 * (counter - self.phase_duration // 2) // (self.phase_duration // 2) + 10
			if counter < half:
				width = counter * 300 // half
			else:
				width = 300
			
		key = str(width) + '^' + str(height)
		surface = self.surfaces.get(key, None)
		if surface == None:
			surface = pygame.Surface((width, height)).convert()
			surface.fill((0, 0, 0))
			surface.set_alpha(180)
			self.surfaces[key] = surface
		
		x = screen.get_width() // 2 - width // 2
		y = 100 - height // 2
		
		screen.blit(surface, (x, y))
		pygame.draw.rect(screen, (255, 255, 255), pygame.Rect(x, y, width, height), 1)
		
	def render(self, screen, counter):
		
		self.playscene.render(screen, counter)
		
		if self.wait_counter > 0: return
		
		self.render_box(screen, self.phase, self.phase_counter)
		
		if self.phase == 'dialog':
			if self.active_portrait != None:
				screen.blit(self.active_portrait, (10, 10))
			
			font_size = 18
			line_height = 23
			last_color = (255, 255, 255)
			y = 70
			
			cursor_height = y + line_height * 3
			
			if self.buffer_clear_counter > 0:
				y -= (10 - self.buffer_clear_counter) * line_height // 10
			for line in self.buffer:
				text = line[0]
				color = line[1]
				last_color = color
				img = get_text(text, font_size, color)
				screen.blit(img, (89, y))
				y += line_height
			
			if self.prompt_for_continue and (counter // 10) % 2 == 0:
				y = cursor_height
				pygame.draw.polygon(screen, last_color, [(300, y), (308, y), (304, y + 5)])

class EndSceneNonStory:
	def __init__(self):
		self.next = self
		
	def process_input(self, events, pressed, axes, mouse):
		for event in events:
			if event.down:
				if event.key == 'start' or event.key == 'spray' or event.key == 'walkie':
					self.next = TransitionScene(self, MainMenuScene())
	
	def update(self, counter):
		pass
	
	def render(self, screen, counter):
		img = get_text("This is the end of the game.", 30, (255, 255, 255))
		img2 = get_text("To see the ending, play in story-mode.", 14, (255, 255, 255))
		
		screen.blit(img, (200 - img.get_width() // 2, 100))
		screen.blit(img2, (200 - img2.get_width() // 2, 150))

class GoodbyeScene:
	def __init__(self):
		self.next = self
		
	def process_input(self, events, pressed, axes, mouse):
		for event in events:
			self.next = None
	
	def update(self, counter):
		pass
	
	def render(self, screen, counter):
		bye = get_text("Goodbye", 16, (255, 255, 255))
		x = (screen.get_width() - bye.get_width()) // 2
		y = (screen.get_height() - bye.get_height()) // 2
		screen.blit(bye, (x, y))
		
		img = get_image('protagonist/wave' + str(((counter // 3) % 4) + 1) + '.png')
		
		screen.blit(img, (192, y + 20))
		

_level_image = {}


class LevelPickerScene:
	def __init__(self):
		self.next = self
		self.x = 0
		self.y = 0
		self.regions = [
			("Sector A", '1-3 2-3 2a-0 3-1'.split()),
			("Sector B", '4-0 5-0 6-0 7-0'.split()),
			("Sector C", '8-0 9-0 10-2 11-0'.split()),
			("Sector D", '12-0 13-0 14-0 14a-0'.split()),
			("Sector E", '15-0 16-0 17-3 18-0'.split()),
			("Sector F", '19a-0 19b-1 19c-0'.split()),
			("Sector G", '19-0 20-0 21-0 21a-0'.split()),
			("Sector H", 'flipmaze 25-0 24-0'.split()),
			("Sector J", '26-0 27-0 28-0'.split()),
			("Sector X", '90-0'.split())
		]
		
	def get_level_image(self, name):
		global _level_image
		img = _level_image.get(name)
		if img == None:
			img = get_image('levelshots/' + name + '.png')
			w = img.get_width()
			h = img.get_height()
			if w > 100:
				h = h * 100 // w
				w = 100
			if h > 60:
				w = w * 60 // h
				h = 60
			img = pygame.transform.scale(img, (w, h))
			_level_image[name] = img
		return img
		
	def process_input(self, events, pressed, axes, mouse):
		for event in events:
			if event.down:
				if event.key == 'left':
					self.x -= 1
					if self.x < 0:
						self.x = 0
					else:
						play_sound('menumove')
						self.y = 0
				elif event.key == 'right':
					self.x += 1
					if self.x > len(self.regions):
						self.x = len(self.regions) - 1
					else:
						play_sound('menumove')
						self.y = 0
				
				elif event.key == 'down':
					self.y += 1
					if self.y >= len(self.regions[self.x][1]):
						self.y -= 1
					else:
						play_sound('menumove')
				elif event.key == 'up':
					self.y -= 1
					if self.y < 0:
						self.y = 0
					else:
						play_sound('menumove')
					
				elif event.key == 'start':
					if self.x == len(self.regions):
						self.next = TransitionScene(self, MainMenuScene())
					else:
						self.next = TransitionScene(self, PlayScene(self.regions[self.x][1][self.y], False))
				
	
	def update(self, counter):
		pass
	
	def render(self, screen, counter):
		screen.fill((48, 48, 48))
		bye = get_text("Level Picker", 18, (255, 255, 255))
		x = (screen.get_width() - bye.get_width()) // 2
		screen.blit(bye, (x, 10))
		
		x = 10
		y = 30
		i = 0
		sector = get_text("Sector: ", 16, (255, 255, 0))
		screen.blit(sector, (x, y))
		
		x += sector.get_width() + 10
		for region in self.regions + [None]:
			
			color = (130, 130, 130)
			if i == self.x:
				color = (255, 255, 255)
			if region == None:
				label = get_text("Main Menu", 16, color)
			else:
				label = get_text(self.regions[i][0].split(' ')[1], 16, color)
			screen.blit(label, (x, y))
			x += label.get_width() + 10
			i += 1
		
		y += sector.get_height() + 10
		
		if self.x < len(self.regions):
			lm = get_level_manager()
			i = 0
			maps = self.regions[self.x][1]
			for m in maps:
				img = self.get_level_image(m)
				screen.blit(img, (30, y))
				
				color = (120, 120, 120)
				if self.y == i:
					color = (255, 255, 255)
				name_label = get_text(lm.get_current_room_name(m), 20, color)
				screen.blit(name_label, (150, y))
				y += 60
				i += 1


class MainMenuScene:
	def __init__(self):
		self.next = self
		self.index = 0
		self.first = True
		has_save_data = get_persisted_forever_string('save_level') != ''
		self.options = [
			('Start New Story Mode', True, lambda x:PlayScene('intro', True)),
			('Resume Saved Game', has_save_data, 'resume_save'),
			('Level Picker', True, lambda x:LevelPickerScene()),
			('Configure Input', True, lambda x:ConfigureInputScene()),
			('Sound Settings', True, lambda x:SettingsScene()),
			('Credits', True, lambda x:CreditsScene(True)),
			('Exit', True, lambda x:GoodbyeScene())
			]

	def process_input(self, events, pressed, axes, mouse):
		go = False
		for event in events:
			if event.down:
				if event.key == 'down':
					self.index += 1
					if self.index < len(self.options) and not self.options[self.index][1]:
						self.index += 1
					if self.index < len(self.options):
						play_sound('menumove')
					else:
						self.index = len(self.options) - 1
				elif event.key == 'up':
					self.index -= 1
					if not self.options[self.index][1]:
						self.index -= 1
					
					if self.index >= 0:
						play_sound('menumove')
					else:
						self.index == 0
				elif event.key == 'start':
					go = True
		
		self.index = max(self.index, 0)
		self.index = min(self.index, len(self.options) - 1)
		
		if go:
			lamb = self.options[self.index][2]
			if lamb == 'resume_save':
				level = get_persistent_state().get_string_forever('save_level')
				next_scene = PlayScene(level, True)
			else:
				next_scene = lamb(None)
			self.next = TransitionScene(self, next_scene)
	
	def update(self, counter):
		get_jukebox().ensure_current_song('title')
		if self.first:
			self.first = False
			get_persistent_state().session = {}
	
	def render(self, screen, counter):
		screen.blit(get_image('misc/title_screen.png'), (0, 0))
		
		title = get_text("Sudo Science", 36, (255, 255, 255))
		w = title.get_width()
		h = title.get_height()
		screen.blit(title, (20, 20))
		
		y = h + 30 + 20
		i = 0
		for option in self.options:
			x = 20
			color = (190, 190, 190)
			if self.index == i:
				color = (255, 255, 255)
				pygame.draw.rect(screen, color, pygame.Rect(x, y + 5, 5, 5))
			elif not option[1]:
				color = (80, 80, 80)
			x = 30
			text = get_text(option[0], 18, color)
			screen.blit(text, (x, y))
			y += text.get_height() + 10
			i += 1

_dark_overlay = None

class PauseScene:
	def __init__(self, playscene):
		global _dark_overlay
		self.next = self
		self.playscene = playscene
		if _dark_overlay == None:
			_dark_overlay = pygame.Surface((400, 300)).convert()
			_dark_overlay.fill((0, 0, 0))
			_dark_overlay.set_alpha(180)
		self.overlay = _dark_overlay
		self.i = 0
		self.options = [
			"Continue",
			"Restart Level",
			"Use Walkie Talkie (Save Game)",
			"Main Menu",
			"Quit"]
		
	
	def process_input(self, events, pressed, axes, mouse):
		for event in events:
			if event.down:
				if event.key == 'up':
					self.i -= 1
					if self.i < 0:
						self.i = 0
					else:
						play_sound('menumove')
				elif event.key == 'down':
					self.i += 1
					if self.i < len(self.options):
						play_sound('menumove')
				elif event.key == 'spray' or event.key == 'start':
					self.do_it()
		if self.i < 0:
			self.i = 0
		if self.i >= len(self.options):
			self.i = len(self.options) - 1
		
	def do_it(self):
		i = max(0, min(len(self.options) - 1, self.i))
		if i == 0:
			# Continue
			self.next = self.playscene
			self.next.next = self.next
		elif i == 1:
			self.next = TransitionScene(self, PlayScene(self.playscene.level.name, self.playscene.story_mode, True))
		elif i == 2:
			self.save_game()
			self.next = DialogScene(self.playscene, 'save')
		elif i == 3:
			self.next = TransitionScene(self, MainMenuScene())
		elif i == 4:
			self.next = None
	
	def save_game(self):
		ps = get_persistent_state()
		increment_persisted_forever_int('research', get_persisted_session_int('research'))
		set_persisted_session_int('research', 0)
		get_persistent_state().set_string_forever('save_level', self.playscene.level.name)
		get_persistent_state().save_game()
	def update(self, counter):
		pass
	
	def render(self, screen, counter):
		self.playscene.render(screen, counter)
		screen.blit(self.overlay, (0, 0))
		y = 100
		
		i = 0
		for option in self.options:
			text = self.options[i]
			color = (100, 100, 100)
			if self.i == i:
				color = (255, 255, 255)
			img = get_text(text, 18, color)
			
			x = screen.get_width() // 2 - img.get_width() // 2
			screen.blit(img, (x, y))
			i += 1
			y += img.get_height() + 20

class PlayScene:
	def __init__(self, level_name, story_mode, temp_suppress=False):
		self.story_mode = story_mode
		persistence_change_level()
		if level_name == '1-3':
			set_persisted_level_int('decontaminant', 3)
		
		self.research_collected = 0
		self.next = self
		self.level_name = level_name
		self.level = Level(level_name, self)
		self.player = Sprite(17, 177, 32, 'main')
		self.sprites = [self.player]
		self.player.level = self.level
		for rat in self.level.rats:
			rat.standingon = None
			self.sprites.append(rat)
		self.overlay = PlaySceneOverlay(self, self.level)
		self.do_not_override_start = False
		level_manager = get_level_manager()
		self.counter = 0
		start = level_manager.get_starting_point_for_level(level_name)
		self.do_stuff = get_hacks_for_level(level_name, 'do_stuff')
		if level_name == '99-0':
			self.player.set_automation(Automation('99-0', '99_p'))
		
		if start != None:
			x = start[0]
			y = start[1]
			z = start[2]
			dir = 's'
			if len(start) >= 4:
				dir = start[3]
			self.do_not_override_start = True
		
			self.player.x = x * 16 + 8
			self.player.y = y * 16 + 8
			self.player.z = z * 8
			self.player.last_direction_of_movement = dir
			self.player.standingon = self.level.get_tile_at(x, y, z - 1)
		
		self.holograms = []
		self.x_offset_override = 0
		level_pixel_width = self.level.height * 16 + self.level.width * 16
		level_pixel_height = level_pixel_width // 2 + 12 * 8
		self.fixed_x = level_pixel_width < 400
		self.fixed_y = level_pixel_height < 300
		if self.level.name == '28-0':
			self.fixed_x = True
			self.fixed_y = True
			self.x_offset_override = 40
		self.target_camera_x = None
		self.target_camera_y = None
		self.camera_x = None
		self.camera_y = None
		if self.do_stuff != None:
			self.do_stuff(self, self.level, -1)
		
		if self.story_mode and not temp_suppress:
			ds = get_startup_dialog(self)
			if ds != None:
				self.next = ds
	
	def process_input(self, events, pressed, axes, mouse):
		for event in events:
			if event.key == 'spray' and event.down and self.player.spray_counter < 0:
				if get_persisted_level_int('decontaminant') > 0:
					play_sound('spray.wav')
					increment_persisted_level_int('decontaminant', -1)
					self.player.spray_counter = 30
					if self.level.spray_from([self.player] + self.holograms):
						if self.level.name == '1-3' and self.story_mode:
							self.next = DialogScene(self, 'a1b')
			if event.key == 'start' and event.down and self.level.name != 'intro' and self.level.name != '99-0':
				self.next = PauseScene(self)
		
		if not self.player.immobilized and self.player.automation == None:
			dx = axes[0]
			dy = axes[1]
			if self.player.spray_counter < 0:
				self.player.dx = dx
				self.player.dy = dy
				
				for h in self.holograms:
					h.dx = dx
					h.dy = dy
		
	
	def update(self, counter):
		level = self.level
		player = self.player
		
		if self.do_stuff != None:
			self.do_stuff(self, level, self.counter)
		
		if player.death_by_rat > 0:
			player.death_by_rat += 1
			if player.death_by_rat == 90:
				self.restart_level()
		
		filtered = []
		for sprite in self.sprites:
			sprite.update(level)
			sprite = sprite.get_replacement_sprite()
			if not sprite.garbage_collect:
				filtered.append(sprite)
			
			if sprite.israt and not player.death_by_rat:
				dx = sprite.x - player.x
				dy = sprite.y - player.y
				distance = dx * dx + dy * dy
				if distance < 144:
					player.death_by_rat = 1
					player.immobilized = True
		
		
		
		self.sprites = filtered + level.get_new_sprites()
		
		if self.player.death_counter == 1:
			self.restart_level()
		
		if self.player.z < -140:
			self.restart_level()
		
		self.level.switch_manager.update_enabled(self.sprites, False)
		
		self.do_hologram_stuff()
		
		self.level.coil_manager.set_fresh_coils(self.level.get_coils())
		self.level.coil_manager.find_zapped_sprites(self.sprites)
		
		if self.level.complete:
			next_level = get_level_manager().get_next_level(self.level.name)
			if next_level == None or (self.level.name == '90-0' and self.story_mode == False):
				self.next = TransitionScene(self, EndSceneNonStory())
			else:
				increment_persisted_session_int('research', get_persisted_level_int('research'))
				set_persisted_level_int('research', 0)
				self.next = TransitionScene(self, PlayScene(next_level, self.story_mode))
		
		sprites_to_add = []
		sprites_to_remove = []
		
		self.level.update(self.sprites, sprites_to_add, sprites_to_remove)

		for sprite in sprites_to_remove:
			sprite.garbage_collect = True
		for sprite in sprites_to_add:
			self.sprites.append(sprite)
		
		get_jukebox().update(self.level.name)
		
		
	def do_hologram_stuff(self):
		self.level.hologram_manager.update(self, self.level, self.player)
		
		anim = self.level.hologram_manager.animation_sequence()
		if anim != None:
			self.player.immobilized = True
			if anim == 0:
				self.player.immobilized = False
		
		for sprite in self.level.hologram_manager.get_new_sprites():
			self.sprites.append(sprite)
			self.holograms.append(sprite)
		self.counter += 1
	
	
	def restart_level(self):
		self.next = TransitionScene(self, PlayScene(self.level.name, self.story_mode, True))
	
	def render(self, screen, counter):
		
		player_position = self.player.pixel_position(0, 0, None)
		player_x = player_position[0]
		player_y = player_position[1]
		
		if self.fixed_x:
			self.target_camera_x = screen.get_width() // 2
		else:
			self.target_camera_x = screen.get_width() // 2 - player_x
		
		if self.fixed_y:
			self.target_camera_y = 50
		else:
			self.target_camera_y = 120 - player_y
		
		if self.camera_x == None:
			self.camera_x = self.target_camera_x
		if self.camera_y == None:
			self.camera_y = self.target_camera_y
		
		max_pan_speed = 4
		if self.camera_x != self.target_camera_x:
			if abs(self.camera_x - self.target_camera_x) == 1:
				self.camera_x = self.target_camera_x
			else:
				new_camera_x = (self.camera_x + self.target_camera_x) // 2
				if abs(new_camera_x - self.camera_x) > max_pan_speed:
					if self.camera_x < new_camera_x:
						self.camera_x += max_pan_speed
					else:
						self.camera_x -= max_pan_speed
				else:
					self.camera_x = new_camera_x
		
		if self.camera_y != self.target_camera_y:
			if abs(self.camera_y - self.target_camera_y) == 1:
				self.camera_y = self.target_camera_y
			else:
				new_camera_y = (self.camera_y + self.target_camera_y) // 2
				if abs(new_camera_y - self.camera_y) > max_pan_speed:
					if self.camera_y < new_camera_y:
						self.camera_y += max_pan_speed
					else:
						self.camera_y -= max_pan_speed
				else:
					self.camera_y = new_camera_y
		
		self.level.render(screen, self.camera_x + self.x_offset_override, self.camera_y, self.sprites, counter)
		
		self.overlay.render(screen, counter)

class PlaySceneOverlay:
	def __init__(self, playscene, level):
		self.playscene = playscene
		self.level = level
	
	def pad_with_zeroes(self, num, digits):
		s = str(num)
		while len(s) < digits:
			s = '0' + s
		return s
	
	def render(self, screen, render_counter):
		bg = get_image('misc/status_overlay.png')
		goo = get_image('tiles/mediumdegoo.png')
		rp = get_image('tiles/researchpapers.png')

		screen.blit(bg, (0, screen.get_height() - bg.get_height()))
		screen.blit(goo, (0, screen.get_height() - goo.get_height() - 14))
		screen.blit(rp, (24, screen.get_height() - rp.get_height() - 0))
		
		research_saved = get_persisted_forever_int('research')
		session_research = get_persisted_session_int('research')
		level_research = get_persisted_level_int('research')
		r = research_saved + level_research + session_research
		
		goo_count = get_text(self.pad_with_zeroes(get_persisted_level_int('decontaminant'), 3), 16, (100, 255, 100))
		rp_count = get_text(self.pad_with_zeroes(r, 3), 16, (255, 255, 255))
		loc = get_level_manager().get_current_room_name(self.level.name)
		if loc == None:
			loc = "Room Needs Name"
		loc = get_text(loc, 12, (255, 255, 255))
		
		screen.blit(goo_count, (30, screen.get_height() - goo.get_height() - 5))
		screen.blit(rp_count, (54, screen.get_height() - rp.get_height() + 11))
		screen.blit(loc, (screen.get_width() - 90, screen.get_height() - loc.get_height() - 4))
		

_set_input_shade = None

def get_set_input_shade():
	global _set_input_shade
	if _set_input_shade == None:
		_set_input_shade = pygame.Surface((400, 300)).convert()
		_set_input_shade.fill((0, 0, 0))
		_set_input_shade.set_alpha(170)
	return _set_input_shade
	
class SetInputScene:
	def __init__(self, action, is_keyboard, prev_scene):
		self.next = self
		self.prev = prev_scene
		self.is_keyboard = is_keyboard
		self.action = action
		font_size = 20
		self.tryanother = get_text("Key in use, try another", 12, (255, 0, 0))
		self.show_tryanother = False
		font_color = (255, 255, 255)
		self.counter = 0
		if self.is_keyboard:
			self.top_img = get_text("Press the key on the keyboard", font_size, font_color)
		else:
			self.top_img = get_text("Press the button on the gamepad", font_size, font_color)
		self.bottom_img = get_text('that will be used to ' +{
			'left': "move left",
			'right': 'move right',
			'up': 'move up',
			'down': 'move down',
			'start': 'pause or confirm menus',
			'spray': 'spray decontaminant',
			'walkie': 'use your walkie talkie (save)'
		}[action], font_size, font_color)
		
	
	def key_in_use(self, code, action):
		km = get_input_manager()._key_mapping
		for k in km.keys():
			if k == code:
				if km[k] != action:
					self.show_tryanother = True
					return True
				else:
					return False
		return False
	
	def leave(self):
		self.next = self.prev
		self.prev.next = self.prev
		
	def process_input(self, events, pressed, axes, mouse):
		im = get_input_manager()
		
		if self.is_keyboard:
			for ku in im.raw_keyups:
				if not self.key_in_use(ku, self.action):
					im.set_key_config(self.action, ku)
					self.leave()
					break
		else:
			# TODO: disable button to get here if self.active_actual_joystick is -1
			js = im.actual_joysticks[im.active_actual_joystick]
			js_config = im.joysticks[im.active_joystick]
			for i in range(js.get_numbuttons()):
				if js.get_button(i):
					js_config[self.action] = ('button', i)
					self.leave()
					return
			for i in range(js.get_numaxes()):
				value = js.get_axis(i)
				if abs(value) > .3:
					sign = '+' if (value > 0) else '-'
					js_config[self.action] = ('axis', i, 'x' + sign)
					self.leave()
					return
			
			for i in range(js.get_numhats()):
				value = js.get_hat(i)
				if value[0] != 0 or value[1] != 0:
					axis = 'x' if (value[1] == 0) else 'y'
					value = value[0] if (axis == 'x') else value[1]
					sign = '+' if (value > 0) else '-'
					js_config[self.action] = ('hat', i, axis + sign)
					self.leave()
					return
	
	def update(self, counter):
		self.counter
	
	def render(self, screen, counter):
		self.prev.render(screen, counter)
		screen.blit(get_set_input_shade(), (0, 0))
		if self.is_keyboard:
			label = "Press the key on the keyboard"
		else:
			label = "Press the button on the gamepad"
		#label += actions[self.action]
		h = self.top_img.get_height()
		y = 150 - h - 3
		left = 25
		top = y - 20
		width = 350
		height = h * 2 + 6 + 40
		pygame.draw.rect(screen, (0, 0, 0), pygame.Rect(left, top, width, height))
		pygame.draw.rect(screen, (128, 128, 128), pygame.Rect(left, top, width, height), 1)
		screen.blit(self.top_img, (200 - self.top_img.get_width() // 2, y))
		screen.blit(self.bottom_img, (200 - self.bottom_img.get_width() // 2, 150 + 3))
		if self.show_tryanother:
			screen.blit(self.tryanother, (200 - self.tryanother.get_width() // 2, top + height - 15))
		

# Sound* settings
class SettingsScene:
	def __init__(self):
		self.next = self
		self.i = 0
		self.jk = get_jukebox()
	
	def toggle_magic(self):
		z = 1 - get_persisted_forever_int('magic')
		set_persisted_forever_int('magic', z)
	
	def process_input(self, events, pressed, axes, mouse):
		for event in events:
			if event.down:
				if event.key == 'up':
					self.i -= 1
					if self.i == -1:
						self.i = 0
					else:
						play_sound('menumove')
				if event.key == 'down':
					self.i += 1
					if self.i > 3:
						self.i = 3
					else:
						play_sound('menumove')
				
				if event.key == 'right':
					if self.i == 0: # SFX
						self.jk.set_sfx_volume(self.jk.get_sfx_volume() + 10)
						play_sound('menumove')
					elif self.i == 1: # Music
						self.jk.set_music_volume(self.jk.get_music_volume() + 10)
						play_sound('menumove')
					elif self.i == 2:
						self.toggle_magic()
						play_sound('menumove')
				
				elif event.key == 'left':
					if self.i == 0: # SFX
						self.jk.set_sfx_volume(self.jk.get_sfx_volume() - 10)
						play_sound('menumove')
					elif self.i == 1: # Music
						self.jk.set_music_volume(self.jk.get_music_volume() - 10)
						play_sound('menumove')
					elif self.i == 2:
						self.toggle_magic()
						play_sound('menumove')
				
				elif event.key == 'start':
					if self.i == 2:
						self.toggle_magic()
						play_sound('menumove')
					elif self.i == 3:
						get_persistent_state().set_int_forever('sfx', self.jk.get_sfx_volume())
						get_persistent_state().set_int_forever('music', self.jk.get_music_volume())
						self.next = TransitionScene(self, MainMenuScene())
						get_persistent_state().save_game()
						play_sound('menumove')
						
				
					
	
	def update(self, counter):
		pass
		
	
	def render(self, screen, counter):
		header = get_text("Sound Settings", 30, (255, 255, 255))
		x = (screen.get_width() - header.get_width()) // 2
		y = 30
		screen.blit(header, (x, y))
		
		y += header.get_height() + 30
		
		g = (120, 120, 120)
		w = (255, 255, 255)
		
		c = w if self.i == 0 else g
		img = get_text("SFX Volume: " + str(self.jk.get_sfx_volume()) + "%", 18, c)
		screen.blit(img, ((screen.get_width() - img.get_width()) // 2, y))
		y += img.get_height() + 40
		
		c = w if self.i == 1 else g
		img = get_text("Music Volume: " + str(self.jk.get_music_volume()) + "%", 18, c)
		screen.blit(img, ((screen.get_width() - img.get_width()) // 2, y))
		y += img.get_height() + 40
		
		c = w if self.i == 2 else g
		t = "More Magic " if (get_persisted_forever_int('magic') == 1) else "Magic"
		img = get_text(t, 18, c)
		screen.blit(img, ((screen.get_width() - img.get_width()) // 2, y))
		y += img.get_height() + 40
		
		c = w if self.i == 3 else g
		img = get_text("Return to Main Menu", 18, c)
		screen.blit(img, ((screen.get_width() - img.get_width()) // 2, y))
		y += img.get_height() + 40
		
		
		

class TitleScene:
	def __init__(self):
		self.next = self
		self.text = get_text(
			"If you can read this, then you are",
			24, (255, 255, 0))
		self.textb = get_text(
			"100% set up to use Python + PyGame",
			24, (255, 255, 0))
		self.x = 0
		self.y = 0

	def process_input(self, events, pressed, axes, mouse):
		pass
	
	def update(self, counter):
		self.x += 1
		self.y += 2
	
	def render(self, screen, counter):
		w = screen.get_width()
		h = screen.get_height()
		self.x = (self.x + w) % w
		self.y = (self.y + h) % h
		
		for t in ((0, 0), (-1, 0), (0, -1), (-1, -1)):
			x = self.x + w * t[0]
			y = self.y + h * t[1]
			screen.blit(self.text, (x, y))
			screen.blit(self.textb, (x, y + self.textb.get_height() + 4))
			

class TransitionScene:
	def __init__(self, before, after):
		self.next = self
		self.before = before
		self.after = after
		self.counter = 0
		self.blackness = None
		
	def process_input(self, events, pressed, axes, mouse):
		pass
	
	def update(self, counter):
		self.counter += 1
		if self.counter >= 60:
			self.next = self.after
		
	def render(self, screen, counter):
		
		if self.counter <= 30:
			opacity = 255 * self.counter // 30
			bg = self.before
		else:
			opacity = (60 - self.counter) * 255 // 30
			bg = self.after
		
		if self.blackness == None:
			self.blackness = pygame.Surface((screen.get_width(), screen.get_height())).convert()
			self.blackness.fill((0, 0, 0))
		
		bg.render(screen, counter)
		self.blackness.set_alpha(opacity)
		screen.blit(self.blackness, (0, 0))

class Automation:
	def __init__(self, level, type):
		self.type = type
		self.level = level
		self.counter = 0
		self.intro_dialog_start = 329 + 14
	
	# (dx, dy)
	def get_next_values(self):
		c = self.counter
		sprite = self.sprite
		self.counter += 1
		o = None
		if self.level == '99-0':
			o = self.do_99_player(self.level, c, sprite)
		else:
			if self.type == 'intro_janitor':
				o = self.do_intro_janitor(self.level, c, sprite)
			elif self.type == 'intro_supervisor':
				o = self.do_intro_supervisor(self.level, c, sprite)
			elif self.type == 'intro_protagonist':
				o = self.do_intro_protagonist(self.level, c, sprite)
			if o == None:
				return (0, 0)
		return o
	
	def do_99_player(self, level, counter, sprite):
		
		if counter < 15:
			return (0, 0)
		elif counter < 23:
			return (-1.5, 0)
		elif counter < 40:
			return (0, 1.5)
		elif counter < 57:
			return (0, -1.5)
		elif counter < 69:
			return (-1.5, 0)
		elif counter == 80:
			
			
			drf2 = get_tile_store().get_tile('drf2')
			sprite.level.modify_block(4, 4, 1, drf2)
		return (0, 0)
	
	def do_intro_janitor(self, level, counter, sprite):
		
		if counter < 124:
			if counter < 8:
				x = 1
			else:
				x = 0
			return (x, 2)
		elif counter < 128:
			return (-2, 0)
		
		
		elif counter > 222 and counter < 312:
			t = counter - 222
			if t < 4:
				return (-1, 0)
			if t == 40:
				sprite.holding_spray = True
			if t == 80:
				sprite.holding_spray = False
			if t > 80 and t < 85:
				return (1, 0)
			elif t == 85:
				return (-1, 0)
		elif counter > 492:
			t = counter - 492
			if t < 4:
				return (-1, 0)
			if t == 20:
				sprite.holding_walkie = True
			if t == 80:
				sprite.holding_walkie = False
			if t > 80 and t < 84:
				return (1, 0)
			if t == 84:
				return (-1, 0)
			
			if t > 100:
				t = t - 100
				if t < 4:
					return (2, 0)
				elif t < 33:
					return (0, 2)
				else:
					sprite.garbage_collect = True
		
			
	
	def do_intro_supervisor(self, level, counter, sprite):
		leave_begin = 373
		if counter < 116:
			if counter < 8:
				x = 1
			else:
				x = 0
			return (x, 2)
		elif counter < 120:
			return (-2, 0)
		
		elif counter > leave_begin:
			t = leave_begin
			if counter < t + 16:
				return (1, 0)
			elif counter < t + 16 + 48:
				return (0, 1)
			elif counter == 450:
				sprite.garbage_collect = True
	
	def do_intro_protagonist(self, level, counter, sprite):
		s = 180
		if counter == s + 132:
			sprite.intro_hack = True
			level.modify_block(int(sprite.x // 16), int(sprite.y // 16), int(sprite.z // 8), get_tile_store().get_tile('54'))
		if counter < s + 132:
			return None
		if counter < s + 148:
			return (0, 1)
		if counter < s + 149:
			return (1, 0)
		

_block_images_for_sprites = None
_surface_cache = {}

def get_surface(width, height):
	global _surface_cache
	k = str(width) + '|' + str(height)
	surface = _surface_cache.get(k, None)
	if surface == None:
		surface = pygame.Surface((width, height)).convert()
		_surface_cache[k] = surface
	return surface
	
def get_teleporter_image(going_out, counter, type):
	counter = min(60, max(0, counter))
	if going_out:
		counter = 60 - counter
	
	if counter < 30:
		ao = 255 - counter * 255 / 30
		bo = 255
	else:
		ao = 0
		bo = (60 - counter) * 255 / 30
	if type == 'main':
		imgs = (
			get_image('protagonist/s.png'),
			get_image('static/character' + str((int(counter // 2) % 4) + 1) + '.png'))
	else:
		block_id = type[len('block|'):]
		if going_out and counter < 6:
			return None
		ts = get_tile_store()
		battery = block_id in('45' or '46')
		img = get_image('static/block' + str((int(counter // 2) % 4) + 1) + '.png')
		if battery:
			img = get_image('static/battery' + str((int(counter // 2) % 4) + 1) + '.png')
		imgs = (
			ts.get_tile(block_id).get_image(counter),
			img)

	return (ao, bo, imgs[0], imgs[1])
	

def copy_surface(surface):
	return pygame.Surface(surface)

def blit_at_opacity(target, source, x, y, opacity):
	t = get_surface(source.get_width(), source.get_height())
	t.blit(screen, (-x, -y))
	t.blit(source)
	t.set_alpha(opacity)
	screen.blit(t, (x, y))

class Sprite:
	# sprite coordinates are assuming the grid is 16x16 tiles
	# these get transposed into pixel coordinates and
	# are converted into tile coords by simply dividing by 16
	def __init__(self, x, y, z, type, level=None):
		global _block_images_for_sprites
		self.garbage_collect = False
		self.x = x + 0.0
		self.y = y + 0.0
		self.z = z
		self.dx = 0
		self.dy = 0
		self.dz = 0
		self.level = None # for intro hack, only
		self.automation = None
		self.falling = False
		self.standingon = None
		self.type = type
		self.immobilized = False
		self.spray_counter = -1
		self.ismain = type == 'main'
		self.main_or_hologram = self.ismain or type == 'hologram|main'
		self.isjanitor = type == 'janitor'
		self.holding_spray = False
		self.clone_creating = False
		self.holding_walkie = False
		self.intro_hack = False
		self.issupervisor = type == 'supervisor'
		self.isblock = type.startswith('block|')
		self.israt = type.startswith('rat|')
		self.death_by_rat = 0
		self.impeded_last_go_round = False
		self.rat_trot_mode = None
		self.rat_blocked_counter = -42
		if self.israt:
			if type.endswith('nw'):
				self.rat_trot_mode = 'nw'
			elif type.endswith('ne'):
				self.rat_trot_mode = 'ne'
			elif type.endswith('sw'):
				self.rat_trot_mode = 'sw'
			else:
				self.rat_trot_mode = 'se'
			self.type = 'rat'
		self.rat_move_max = 27
		self.rat_move_counter = self.rat_move_max
		
		self.death_counter = -1
		self.death_type = None
		self.height = 4
		if self.isblock:
			self.height = 2
		self.tsend = type.startswith('teleport|')
		self.trecv = type.startswith('receiving|')
		self.staticy = self.tsend or self.trecv
		self.ttype = None
		if self.staticy:
			self.ttype = '|'.join(type.split('|')[1:])
		self.ttl = None
		if self.staticy:
			self.ttl = 60
		self.last_direction_of_movement = 's'
		self.direction_queue = ['s']
		self.is_moving = False
		self.pushing = None
		if self.isblock:
			self.height = 2
		
		if _block_images_for_sprites == None:
			tile_store = get_tile_store()
			ids = tile_store.get_all_block_tiles()
			_block_images_for_sprites = {}
			for id in ids:
				_block_images_for_sprites[id] = tile_store.get_tile(id)
		
		if self.isblock:
			self.block_id = type.split('|')[-1] # should be the tile ID
			self.block_tile = get_tile_store().get_tile(self.block_id)
		
	
	def set_automation(self, automation):
		self.automation = automation
		automation.sprite = self
	
	
	def render_me(self, screen, xOffset, yOffset, render_counter):
		img = self.get_image(render_counter)
		if img != None:
			coords = self.pixel_position(xOffset, yOffset, img)
			
			if self.staticy:
				things = get_teleporter_image(self.tsend, self.ttl, self.ttype)
				if things == None: return
				ao = things[0]
				bo = things[1]
				ai = things[2]
				bi = things[3]
				
				for x in ((bo, bi), (ao, ai)):
					o = x[0]
					img = x[1]
					
					if o > 0:
						t = get_surface(img.get_width(), img.get_height())
						t.blit(screen, (-coords[0], -coords[1]))
						t.blit(img, (0, 0))
						t.set_alpha(o)
						screen.blit(t, (coords[0], coords[1]))	
			else:
				screen.blit(img, coords)
			
		return (self, xOffset, yOffset, render_counter)
		
	
	
	def get_image(self, render_counter):
		img = None
		if self.main_or_hologram:
			if not self.ismain:
				if self.clone_creating:
					return get_image('scan/character'+str(((render_counter // 2) % 4) + 1)+'.png')
				if render_counter % 2 == 1:
					return None #flickering holograms
			else:
				if self.level != None and self.level.name == 'intro':
					if self.intro_hack:
						pass
					else:
						return get_image('protagonist/nwsit.png')
			if self.death_counter > 0:
				path = 'protagonist/s.png'
				if self.death_type == 'goo':
					path = 'protagonist/goo' + str((int(render_counter // 3) % 4) + 1) + '.png'
				elif self.death_type == 'bazat':
					path = 'protagonist/zap' + str((int(render_counter // 3) % 4) + 1) + '.png'
			elif self.death_by_rat > 0:
				path = 'protagonist/rat'
				path += str((int(self.death_by_rat // 3) % 4) + 1) + '.png'
			elif self.standingon == None:
				path = 'protagonist/fall' + str((int(render_counter // 3) % 4) + 1) + '.png'
			else:
				dir = self.last_direction_of_movement
				if self.pushing != None:
					dir = self.pushing.lower() + 'push'
				path = 'protagonist/' + dir
				if self.spray_counter >= 0:
					path += 'spray' + str(((render_counter // 3) % 4) + 1)
				elif self.is_moving:
					path += str([1, 2, 3, 4, 3, 2][(render_counter // 6) % 6])
				path += '.png'
			img = get_image(path)
		elif self.israt:
			path = 'rat/'
			if self.death_type == 'bazat':
				path += 'zap' + str(((render_counter // 3) % 4) + 1)
			else:
				path += self.rat_trot_mode
				if self.is_moving:
					path += str([1, 2, 3, 4, 3, 2][(render_counter // 6) % 6])
			path += '.png'
			img = get_image(path)
		elif self.isjanitor:
			if self.holding_spray:
				path = 'janitor/spray2.png'
			elif self.holding_walkie:
				path = 'janitor/walkietalkie2.png'
			else:
				dir = self.last_direction_of_movement
				if dir == 's' or dir == 'n':
					dir += 'e'
				elif dir == 'e' or dir == 'w':
					dir = 's' + dir
				path = 'janitor/' + dir
				if self.is_moving:
					path += str([1, 2, 3, 4, 3, 2][(render_counter // 6) % 6])
				path += '.png'
			img = get_image(path)
		elif self.issupervisor:
			dir = self.last_direction_of_movement
			#TODO: get n, s, e, w images of supervisor|janitor
			if dir == 's' or dir == 'n':
				dir += 'e'
			elif dir == 'e' or dir == 'w':
				dir = 's' + dir
			path = 'supervisor/' + dir
			if self.is_moving:
				path += str([1, 2, 3, 4, 3, 2][(render_counter // 6) % 6])
			path += '.png'
			img = get_image(path)
		elif self.staticy:
			img = get_image('static/character' + str(((render_counter // 6) % 4) + 1) + '.png')
		elif self.isblock:
			img = self.block_tile.get_image(render_counter)

		return img
	
	def pixel_position(self, xOffset, yOffset, img):
		x = self.x - self.y
		y = (self.x + self.y) / 2
		if img == None:
			w = 16
			h = 32
		else:
			w = img.get_width()
			h = img.get_height()
		x = x - w / 2
		y = y - self.z - h + 8
		if self.isblock:
			y += 8
		
		if self.main_or_hologram or self.staticy or self.isjanitor or self.issupervisor or self.israt:
			if self.staticy and self.ttype.startswith('block|'):
				y += 16
			y += 8

		platform = self.standingon
		if platform != None and platform.stairs:
			left = platform.topography[3] * 8
			right = platform.topography[1] * 8
			if platform.entrance == 'SW' or platform.entrance == 'NE':
				ap = self.y % 16 + 0.0
				p = 16.0 - ap
			else:
				p = self.x % 16 + 0.0
				ap = 16.0 - p
			dy = int((left * ap + p * right) / 16.0)
			y -= dy
			
		output = (int(x + xOffset), int(y + yOffset))
		
		return output
	
	def debug_stats(self):
		return ' '.join(safe_map(str, [
			'x:', self.x,
			'y:', self.y,
			'z:', self.z,
			'dx:', self.dx,
			'dy:', self.dy,
			'dz:', self.dz]))
		
	
	def get_replacement_sprite(self):
		if self.trecv and self.garbage_collect:
			self.prototype.garbage_collect = False
			return self.prototype
		return self
	
	def flip_rat_trot_mode(self):
		rtm = self.rat_trot_mode
		if rtm == 'nw':
			self.rat_trot_mode = 'se'
		elif rtm == 'se':
			self.rat_trot_mode = 'nw'
		elif rtm == 'ne':
			self.rat_trot_mode = 'sw'
		else:
			self.rat_trot_mode = 'ne'
	
	
	def move_rat(self, level):
		self.rat_blocked_counter -= 1
		rtm = self.rat_trot_mode
		if self.rat_blocked_counter == 0:
			self.flip_rat_trot_mode()
		
		if self.z < -200:
			self.garbage_collect = True
		
		if self.rat_blocked_counter < 0:
			if self.impeded_last_go_round:
				self.rat_blocked_counter = 60
			else:
				v = 0.8
				if rtm == 'nw':
					self.dx = -v
				elif rtm == 'ne':
					self.dy = -v
				elif rtm == 'sw':
					self.dy = v
				else:
					self.dx = v
				
				newx = self.x + self.dx
				newy = self.y + self.dy
				col = int(newx // 16)
				row = int(newy // 16)
				layer = int(self.z // 8) - 1
				if layer >= 0:
					tile = level.get_tile_at(col, row, layer)
				else:
					tile = None
					
				if tile == None or not tile.blocking:
					self.rat_blocked_counter = 60
					self.dx = 0
					self.dy = 0
	
	def update(self, level):
		self.death_counter -= 1
		self.spray_counter -= 1
		
		if self.israt:
			self.move_rat(level)
		
		if self.death_counter == 1:
			self.garbage_collect = True
		
		if self.automation != None:
			auto_xy = self.automation.get_next_values()
			self.dx += auto_xy[0]
			self.dy += auto_xy[1]
		
		self.pushing = None
		self.falling = False
		if self.ttl != None:
			self.ttl -= 1
			if self.ttl <= 0:
				self.garbage_collect = True
				if self.trecv:
					self.prototype.x = self.x
					self.prototype.y = self.y
					self.prototype.z = self.z
					self.prototype.is_moving = False
					self.prototype.immobilized = False
		
		if self.immobilized:
			return
		if self.standingon == None:
			self.dz = -1

		if self.dz != 0:
			platform_data = level.get_platform_below(int(self.x) // 16, int(self.y) // 16, self.z, True)
			if platform_data != None:
				z = platform_data[0]
				platform = platform_data[1]
				if z < self.dz + self.z:
					self.z += self.dz
					self.falling = True
				else:
					self.z = z
					self.standingon = platform
			else:
				self.standingon = None
				self.z += self.dz
		layer = self.z // 8
		on_new_coordinates_now = False
		new_platform = self.standingon
		
		starting_col = int(self.x) // 16
		starting_row = int(self.y) // 16
		
		if self.standingon != None:
			if self.standingon.teleporter and not self.staticy:
				destination = level.teleporters.get_destination(starting_col, starting_row, int(self.z // 8) - 1)
				
				if destination != None:
					if destination == 'blocked':
						pass
						# TODO: play blocked sound
					else:
						level.teleporters.teleport_sprite(self, destination)
		
		if self.dz == 0:
			
			ending_col = int(self.x + self.dx) // 16
			ending_row = int(self.y + self.dy) // 16
			
			check_these = []
			if starting_col == ending_col and starting_row == ending_row:
				# if we were cool with it before, we're cool with it now
				direction = None
				opposite = None
			elif starting_col != ending_col and starting_row != ending_row:
				# player moved diagonally to another tile and skipped over intermediate tiles
				# check one of the tiles between to make sure there's a path.
				check_these.append((ending_col, starting_row))
				check_these.append((ending_col, ending_row))
				if starting_col < ending_col:
					if starting_row < ending_row:
						direction = 'S'
						opposite = 'N'
					else:
						direction = 'E'
						opposite = 'W'
				else:
					if starting_row < ending_row:
						direction = 'W'
						opposite = 'E'
					else:
						direction = 'N'
						opposite = 'S'
			else:
				# player moved to a tile cardinally next to it
				check_these.append((ending_col, ending_row))
				if starting_col == ending_col:
					if starting_row < ending_row:
						direction = 'SW'
						opposite = 'NE'
					else:
						direction = 'NE'
						opposite = 'SW'
				else:
					if starting_col < ending_col:
						direction = 'SE'
						opposite = 'NW'
					else:
						direction = 'NW'
						opposite = 'SE'
			
			blocked = False
			topToBottom = True
			clearance = safe_range(self.height)[:]
			if topToBottom:
				clearance = clearance[::-1]
			cellLookup = level.cellLookup
			tilestack = level.grid
			for check in check_these:
				col = check[0]
				row = check[1]
				if col < 0 or col >= level.width or row < 0 or row >= level.height:
					blocked = True
					self.impeded_last_go_round = True
					break
				lookup = cellLookup[col][row]
				tiles = tilestack[col][row]
				for c in clearance:
					if len(lookup) > c + layer and lookup[c + layer] != None:
						tz = lookup[c + layer]
						t = tiles[tz]
						
						if t.blocking:
							if self.main_or_hologram:
								pt = level.get_push_tracker(self)
								prev_push_target = pt.push_target
								pt.push_target = None
								
								if self.ismain and t.isexit:
									level.complete = True
								
								if t.pushable and direction != None and len(direction) == 2:
									blocked = True
									push_key = str(col) + '^' + str(row) + '^' + str(tz)
									pt.push_target = push_key

									if direction == 'NW':
										tcol = col - 1
										trow = row
									elif direction == 'NE':
										tcol = col
										trow = row - 1
									elif direction == 'SW':
										tcol = col
										trow = row + 1
									elif direction == 'SE':
										tcol = col + 1
										trow = row
									else:
										assertion("ERROR: bad direction while pushing block.")
									
									self.pushing = direction
									
									if push_key == prev_push_target:
										pt.push_counter -= 1
									else:
										pt.push_counter = pt.max_push_counter
									if pt.push_counter == 0:
										# Try to do the push
										if tcol >= 0 and tcol < level.width and trow >= 0 and trow < level.height:
											tlookup = level.cellLookup[tcol][trow]
											tstack = level.grid[tcol][trow]
											bottom_free = True
											top_free = True
											bottom_index = c + layer - 1
											top_index = c + layer
											
											if len(tlookup) > bottom_index:
												bottom_index = tlookup[bottom_index]
												if bottom_index != None:
													if tstack[bottom_index].blocking or tstack[bottom_index].cant_push_over:
														bottom_free = False
											
											if len(tlookup) > top_index:
												top_index = tlookup[top_index]
												if top_index != None:
													if tstack[top_index].blocking or tstack[top_index].cant_push_over:
														top_free = False
											
											if bottom_free and top_free:
												# TODO: make sure you're not pushing onto a ramp
												target_level = c + layer - 1
												while len(tlookup) <= target_level:
													tstack.append(None)
													tlookup.append(None)
												_t = tlookup[target_level - 1]
												_i = target_level - 1
												while _i >= 0 and _t == None:
													_t = tlookup[_i]
													_i -= 1
												
												if _i == -1:
													_t = -1
												
												valid = True
												while _t >= 0:
													landing_tile = tstack[_t]
													
													if landing_tile != None:
														if landing_tile.stairs:
															valid = False
														elif landing_tile.no_blocks:
															valid = False
														else:
															valid = True
														break
													_t -= 1
												
												if valid:
													level.push_block(self, col, row, tcol, trow, target_level)
												else:
													pt.push_counter = -1
													pt.push_target = None
											else:
												pt.push_counter = -1
												pt.push_target = None
											break
										else:
											pt.push_counter = -1
											pt.push_target = None
									
								
							# The target tile will always be last in check_these
							if blocked == False and t.stairs and t.entrance == opposite and check == check_these[-1]:
								if c == 0:
									# not blocked
									zbefore = self.z
									self.z += t.height * 8
									break
							else:
								blocked = True
								break
				if blocked:
					break
			
			
			if not blocked:
				old_col = int(self.x // 16)
				old_row = int(self.y // 16)
				self.x += self.dx
				self.y += self.dy
				self.impeded_last_go_round = False
				new_col = int(self.x // 16)
				new_row = int(self.y // 16)
				if old_col != new_col or old_row != new_row:
					on_new_coordinates_now = True
				
				new_layer = int(self.z // 8)
				
				occupying = level.get_tile_at(new_col, new_row, new_layer)
				if occupying != None and occupying.is_goo and self.main_or_hologram:
					self.death_counter = 60
					self.death_type = 'goo'
					self.immobilized = True
					play_sound("irradiated.wav")
			self.impeded_last_go_round = blocked
					
			omg_hax = self.dx != 0 or self.dy != 0
			self.is_moving = omg_hax
			if self.automation == None:
				omg_hax = self.dx != 0 and self.dy != 0
			
				
			if self.dx == 0 and self.dy == 0:
				self.direction_queue = []
			
			
			if omg_hax:
				distance = (self.dx * self.dx + self.dy * self.dy) ** .5
				ndx = self.dx / distance
				ndy = self.dy / distance
				
				tolerance = .35
				dy_off = ndy > -tolerance and ndy < tolerance
				dx_off = ndx < tolerance and ndx > -tolerance
				
				d = None
				
				if dx_off:
					if dy_off:
						pass
					elif ndy < 0:
						d = 'ne'
					else:
						d = 'sw'
				elif ndx < 0:
					if dy_off:
						d = 'nw'
					elif ndy < 0:
						d = 'n'
					else:
						d = 'w'
				else:
					if dy_off:
						d = 'se'
					elif ndy < 0:
						d = 'e'
					else:
						d = 's'
				if d != None:
					self.last_direction_of_movement = d
					self.direction_queue = [d] + self.direction_queue[:4]
					if len(self.direction_queue) == 5:
						a,b,c,d,e = self.direction_queue
						if len(a) == 1 and d == e and len(d) == len(e) and a in d and ((b == c and (b == a or b == d)) or (b == a and c == d)): # BWAHAHAHAHAHA
							self.last_direction_of_movement = d
					else:
						
						for foo in self.direction_queue:
							if len(foo) == 2:
								self.last_direction_of_movement = foo
								break
				
			
		if new_platform != None and new_platform.stairs and on_new_coordinates_now:
			if direction == new_platform.entrance:
				self.z -= new_platform.height * 8
		
		if self.dz == 0:
			col = int(self.x // 16)
			row = int(self.y // 16)
			layer = int(self.z - 1) // 8
			lookup = level.cellLookup[col][row]
			if layer < len(lookup):
				_t = lookup[layer]
				if _t == None:
					self.standingon = None
				else:
					tile = level.grid[col][row][_t]
					if tile.blocking:
						self.standingon = tile
					else:
						self.standingon = None
			else:
				self.standingon = None
		
		if self.isblock and self.standingon != None:
			# turn me back into a real block
			self.garbage_collect = True
			level.modify_block(int(self.x // 16), int(self.y // 16), int(self.z // 8), self.block_tile)
		
		self.dx = 0
		self.dy = 0
		self.dz = 0
		
		if self.main_or_hologram:
			self.try_pick_up_powerups(level)
	
	def try_pick_up_powerups(self, level):
		col = int(self.x // 16)
		row = int(self.y // 16)
		layer = int(self.z // 8)
		if layer < 0: return
		tile = level.get_tile_at(col, row, layer)
		if tile != None and tile.powerup:
			play_sound('pickup.wav')
			level.modify_block(col, row, layer, None)
			if tile.goo:
				increment_persisted_level_int('decontaminant', tile.goo_size)
			else:
				increment_persisted_level_int('research', 1)
				

def trim(string):
	while len(string) > 0 and string[0] in ' \n\r\t':
		string = string[1:]
	while len(string) > 0 and string[-1] in ' \r\n\t':
		string = string[:-1]
	return string

def canonicalize_path(path):
	return path.replace('\\', os.sep).replace('/', os.sep)

def read_file(path):
	path = canonicalize_path(path)
	if os.path.exists(path):
		c = open(canonicalize_path(path), 'rt')
		t = c.read()
		c.close()
		return t
	return None
	
def write_file(path, contents):
	c = open(path, 'wt')
	c.write(contents)
	c.close()

def make_grid(width, height, defaultValue):
	cols = []
	x = 0
	while x < width:
		col = []
		y = 0
		while y < height:
			col.append(defaultValue)
			y += 1
		cols.append(col)
		x += 1
	return cols

def safe_map(function, list):
	i = 0
	l = len(list)
	output = []
	while i < l:
		output.append(function(list[i]))
		i += 1
	return output

def max(a, b):
	if a < b:
		return b
	return a
	
def min(a, b):
	if a < b:
		return a
	return b

_range_4 = [0, 1, 2, 3]
_range_2 = [0, 1]
def safe_range(x):
	global _range_4, _range_2
	if x < 10:
		if x == 4:
			return _range_4
		if x == 2:
			return _range_2
	
	output = [0] * x
	i = 1
	while i < x:
		output[i] = i
		i += 1
	return output

def safe_sorted(list, compare):
	return safe_sorted_helper(list, compare)
	
def safe_sorted_helper(list, compare):
	if len(list) <= 1:
		return list
	if len(list) == 2:
		if compare(list[0], list[1]):
			return list
		else:
			return list[::-1]
	pivot_i = len(list) // 2
	pivot = list[pivot_i]
	list = list[:pivot_i] + list[pivot_i + 1:]
	left = []
	right = []
	for item in list:
		if compare(item, pivot):
			left.append(item)
		else:
			right.append(item)
	
	left = safe_sorted_helper(left, compare)
	left.append(pivot)
	right = safe_sorted_helper(right, compare)
	return left + right

_assertions_break = True
def assertion(message):
	global _assertions_break
	print(message)
	if _assertions_break:
		destroy = 1 / 0

def copy_array(target, source):
	while len(target) > len(source):
		target.pop()
	while len(target) < len(source):
		target.append(None)
	
	i = 0
	while i < len(source):
		target[i] = source[i]
		i += 1
def debug_dict(dictionary):
	return '\n'.join(safe_map(lambda k:str(k) + ": " + str(dictionary[k]), dictionary.keys()))

def debug_list(list):
	return '\n'.join(safe_map(str, list))

def parseInt(string):
	string = trim(string)
	flip = 1
	if len(string) > 0 and string[0] == '-':
		flip = -1
		string = string[1:]
	for c in string:
		if not c in '0123456789':
			return 0
	return flip * int(string)
	
def abs(num):
	if num < 0: return -num
	return num

_debug_message = None
def set_user_debug_message(text):
	global _debug_message
	_debug_message = text

def get_user_debug_message():
	global _debug_message
	return _debug_message			

def main():

	pygame.init()
	pygame.display.set_caption("Sudo Science")
	pygame.display.set_icon(pygame.image.load('icon.png'))
	real_screen = pygame.display.set_mode((800, 600))
	fake_screen = pygame.Surface((400, 300))
	fps = 60

	pressed = {
		'start': False,
		'left': False,
		'right': False,
		'down': False,
		'up': False,
		'action': False
	}
	
	load_persistent_state()
	
	if os.path.exists('start.txt'):
		lines = trim(read_file('start.txt').split('\n'))
		
		if lines[0] == 'normal':
			active_scene = MainMenuScene()
		else:
			level_name = lines[0]
			active_scene = PlayScene(level_name, False)
			if not active_scene.do_not_override_start:
				coords = safe_map(int, lines[1].split(','))
				x = coords[0]
				y = coords[1]
				z = 8
				active_scene.player.x = x * 16 + 8
				active_scene.player.y = y * 16 + 8
				active_scene.player.z = z * 8
			
	else:
		active_scene = MainMenuScene()
	counter = 0
	
	input_manager = get_input_manager()
	
	while active_scene != None:
		start = time.time()
		
		counter += 1
		event_list = []
		event_list = input_manager.get_events()
		pressed = input_manager.my_pressed
		try_quit = input_manager.quitAttempt
		axes = input_manager.axes
		mouse_events = input_manager.get_mouse_events()
		
		active_scene.process_input(event_list, pressed, axes, mouse_events)
		active_scene.update(counter)
		
		fake_screen.fill((0, 0, 0))
		active_scene.render(fake_screen, counter)
		
		debug_message = get_user_debug_message()
		if debug_message != None:
			fake_screen.blit(get_text(debug_message, 20, (255, 0, 0)), (10, 10))
		
		pygame.transform.scale(fake_screen, (real_screen.get_width(), real_screen.get_height()), real_screen)
		
		active_scene = active_scene.next
		
		if try_quit:
			active_scene = None
			
		pygame.display.flip()
		
		end = time.time()
		
		diff = end - start
		if diff == 0:
			rate = 'inf'
		else:
			rate = 1.0 / diff
		delay = 1.0 / fps - diff
		if delay > 0:
			time.sleep(delay)


main()